"""Statistics that are used in hnet."""

import df2onehot as df2onehot
from ismember import ismember

import numpy as np
import pandas as pd
from scipy.stats import hypergeom, ranksums
import statsmodels.stats.multitest as multitest
from sklearn.preprocessing import MinMaxScaler

import os
import tempfile
from pathlib import Path
import itertools


# %% Compute significance
def _compute_significance(df, y, dtypes, specificity=None, verbose=3):
    out=[]
    # Run over all columns
    for i in range(0, df.shape[1]):
        if (i>0) and (verbose>=3): print('')
        if verbose>=3: print('[hnet] >Analyzing [%s] %s' %(dtypes[i], df.columns[i]), end='')
        colname = df.columns[i]

        # Clean nan fields
        [datac, yc] = _nancleaning(df[colname], y)
        # In a two class model, remove 0-catagory
        uiy = np.unique(yc)
        # No need to compute _other_ because it is a mixed group that is auto generated based on y_min
        uiy=uiy[uiy!='_other_']

        if len(uiy)==1 and (uiy=='0'):
            if verbose>=4: print('[hnet] >The response variable [y] has only one catagory; [0] which is seen as the negative class and thus ignored.')
            uiy=uiy[uiy!='0']

        if len(uiy)==2:
            if verbose>=4: print('[hnet] >The response variable [y] has two catagories, the catagory 0 is seen as the negative class and thus ignored.')
            uiy=uiy[uiy!='0']
            uiy=uiy[uiy!='False']
            uiy=uiy[uiy!='false']

        # Run over all target values
        for j in range(0, len(uiy)):
            target = uiy[j]
            # Catagorical
            if (dtypes[i]=='cat'):
                datacOnehot = pd.get_dummies(datac)
                # Remove background column if 2 columns of which 0.0 also there
                if (datacOnehot.shape[1]==2) & (np.any(np.isin(datacOnehot.columns, '0.0'))):
                    datacOnehot.drop(labels=['0.0'], axis=1, inplace=True)
                # Remove False background column if 2 columns and False
                if (datacOnehot.shape[1]==2) & (np.any(np.isin(datacOnehot.columns, False))):
                    datacOnehot.drop(labels=[False], axis=1, inplace=True)
                # Run over all unique entities/cats in column for target vlue
                for k in range(0, datacOnehot.shape[1]):
                    outtest = _prob_hypergeo(datacOnehot.iloc[:, k], yc==target)
                    outtest.update({'y': target})
                    outtest.update({'category_name': colname})
                    out.append(outtest)
            elif dtypes[i]=='num':
                # Numerical
                outtest = _prob_ranksums(datac, yc==target, specificity=specificity)
                outtest.update({'y': target})
                outtest.update({'category_name': colname})
                out.append(outtest)
            else:
                if verbose>=2: print('[hnet] >Warning: Can not process dtype: [%s] <skipping>' %(dtypes[i]))
                # raise ValueError('[hnet] >dtype can not be of type [%s]' %(dtypes[i]))
            # Print dots
            if verbose>=3: print('.', end='')

    if verbose>=3: print('')
    return(out)


# %% Wilcoxon Ranksum test
def _prob_ranksums(datac, yc, specificity=None):
    P=np.nan
    zscore=np.nan
    datac=datac.values
    getsign=''

    # Wilcoxon Ranksum test
    if sum(yc==True)>1 and sum(yc==False)>1:
        [zscore, P] = ranksums(datac[yc==True], datac[yc==False])

    # Store
    out=dict()
    out['P']=P
    out['logP']=np.log(P)
    out['zscore']=zscore
    out['popsize_M']=len(yc)
    out['nr_succes_pop_n']=np.sum(yc==True)
    out['nr_not_succes_pop_n']=np.sum(yc==False)
    out['dtype']='numerical'

    if np.isnan(zscore) is False and np.sign(zscore)>0:
        getsign='high_'
    else:
        getsign='low_'

    if specificity=='low':
        out['category_label']=getsign[:-1]
    elif specificity=='medium':
        out['category_label']=getsign + str(('%.1f' %(np.median(datac[yc==True]))))
    elif specificity=='high':
        out['category_label']=getsign + str(('%.3f' %(np.median(datac[yc==True]))))
    else:
        out['category_label']=''

    return(out)


# %% Hypergeometric test
def _prob_hypergeo(datac, yc):
    """Compute hypergeometric Pvalue.

    Description
    -----------
    Suppose you have a lot of 100 floppy disks (M), and you know that 20 of them are defective (n).
    What is the prbability of drawing zero to 2 floppy disks (N=2), if you select 10 at random (N).
    P=hypergeom.sf(2,100,20,10)

    """
    P = np.nan
    logP = np.nan
    M = len(yc)  # Population size: Total number of samples, eg total number of genes; 10000
    n = np.sum(datac)  # Number of successes in population, known in pathway, eg 2000
    N = np.sum(yc)  # sample size: Random variate, eg clustersize or groupsize, over expressed genes, eg 300
    X = np.sum(np.logical_and(yc, datac)) - 1  # Let op, de -1 is belangrijk omdatje P<X wilt weten ipv P<=X. Als je P<=X doet dan kan je vele false positives krijgen als bijvoorbeeld X=1 en n=1 oid

    # Test
    if np.any(yc) and (X>0):
        P = hypergeom.sf(X, M, n, N)
        logP = hypergeom.logsf(X, M, n, N)

    # Store
    out=dict()
    out['category_label']=datac.name
    out['P']=P
    out['logP']=logP
    out['overlap_X']=X
    out['popsize_M']=M
    out['nr_succes_pop_n']=n
    out['samplesize_N']=N
    out['dtype']='categorical'

    return(out)


# %% Make logscale
def _logscale(simmat_padj):
    # Set minimum amount
    if np.any(~np.isnan(simmat_padj.values.ravel())):
        simmat_padj[simmat_padj==0]=1e-323
        adjmatLog=(-np.log10(simmat_padj)).copy()
        adjmatLog[adjmatLog == -np.inf] = np.nanmax(adjmatLog[adjmatLog != np.inf])
        adjmatLog[adjmatLog == np.inf] = np.nanmax(adjmatLog[adjmatLog != np.inf])
        adjmatLog[adjmatLog == -0] = 0
    else:
        adjmatLog = simmat_padj.copy()
    return(adjmatLog)


# %% Do multiple test correction
def _multipletestcorrectionAdjmat(adjmat, multtest, verbose=3):
    if verbose>=3: print('[hnet] >Multiple test correction using %s' %(multtest))
    # Multiple test correction
    if multtest is not None:
        # Make big row with all pvalues
        tmpP = adjmat.values.ravel()
        # Find not nans
        Iloc = ~np.isnan(tmpP)
        Padj = np.zeros_like(tmpP) * np.nan
        # Do multiple test correction on only the tested ones
        if np.any(Iloc):
            Padj[Iloc]=multitest.multipletests(tmpP[Iloc], method=multtest)[1]
        # Rebuild adjmatrix
        adjmat = pd.DataFrame(data=Padj.reshape(adjmat.shape), columns=adjmat.columns, index=adjmat.index)

    # Return
    return(adjmat)


# %% Do multiple test correction
def _multipletestcorrection(out, multtest, verbose=3):
    # Always do a multiple test correction but do not use it in the filtering step if not desired
    if verbose>=3: print('[hnet] >Multiple test correction using %s' %(multtest))

    if out!=[]:
        # Get pvalues
        Praw = np.array(list(map(lambda x: x['P'], out)))
        Iloc = np.isnan(Praw)
        Praw[Iloc] = 1

        # Multiple test correction
        if multtest is None:
            Padj = Praw
        else:
            # Padj=np.zeros_like(Praw)*np.nan
            Padj=multitest.multipletests(Praw, method=multtest)[1]

        for i in range(0, len(out)):
            out[i].update({'Padj': Padj[i]})

    return(out)


# %% Add combinations
def _make_n_combinations(Xhot, Xlabx, combK, y_min, verbose=3):
    Xlabo=Xlabx.copy()
    if isinstance(y_min, type(None)): y_min=1
    # If any, run over combinations
    if not isinstance(combK, type(None)) and combK>1:
        out_hot = Xhot
        out_labo = Xlabo
        out_labx = list(map(lambda x: [x], Xlabx))
        # Run over all combinations
        for k in np.arange(2, combK + 1):
            # Make smart combinations because of mutual exclusive classes
            [cmbn_hot, cmbn_labX, cmbn_labH, cmbn_labO] = _cmbnN(Xhot, Xlabx, y_min, k)
            # If any combinations is found, add to dataframe
            if len(cmbn_labX)>0:
                if verbose>=3: print('[hnet] >Adding %d none mutual exclusive combinations with k=[%d] features.' %(cmbn_hot.shape[1], k))
                out_hot = pd.concat([out_hot, pd.DataFrame(data=cmbn_hot, columns=cmbn_labH).astype(int)], axis=1)
                out_labo = np.append(out_labo, cmbn_labO, axis=0)
                out_labx = out_labx + cmbn_labX
            else:
                if verbose>=3: print('[hnet] >No combinatorial features detected with k=[%d] features. No need to search for higher k.' %(k))
                break

        # Add to one-hot dataframe
        Xhot = out_hot
        Xlabo = out_labo
        Xlabx = out_labx

    assert Xhot.shape[1]==len(Xlabx), print('one-hot matrix should have equal size with xlabels')
    assert Xhot.shape[1]==len(Xlabo), print('one-hot matrix should have equal size with olabels')
    return(Xhot, Xlabx, Xlabo)


# %% Add combinations
def _cmbnN(Xhot, Xlabx, y_min, k):
    # Take only those varibles if combinations is larger then N (otherwise it is not mutually exclusive)
    [uilabx, uicount]=np.unique(Xlabx, return_counts=True)
    Iloc = np.isin(Xlabx, uilabx[uicount>k])

    # cmnb_labx = np.array(list(itertools.combinations(Xhot.columns[I], k)))
    cmbn_idx = np.array(list(itertools.combinations(np.where(Iloc)[0], k)))
    cmbn_hot = []
    cmbn_labH = []
    cmbn_labX = []
    cmbn_labO = []
    for idx in cmbn_idx:
        # Compute product
        prodFeat = Xhot.iloc[:, idx].prod(axis=1)
        # Store if allowed
        if sum(prodFeat)>=y_min:
            cmbn_hot.append(prodFeat.values)
            cmbn_labH.append('_&_'.join(Xhot.columns[idx]))
            cmbn_labX.append(list(np.unique(Xlabx[idx])))
            cmbn_labO.append('_&_'.join(np.unique(Xlabx[idx])))

    # Make array
    cmbn_hot=np.array(cmbn_hot).T
    # Combine repetative values
    # assert cmbn_hot.shape[1]==len(cmbn_labX), print('one-hot matrix should have equal size with labels')
    return(cmbn_hot, cmbn_labX, cmbn_labH, cmbn_labO)


# %% Add columns
def _addcolumns(simmat_padj, colnames, Xlabx, catnames):
    Iloc = np.isin(colnames.values.astype(str), simmat_padj.index.values)
    if np.any(Iloc):
        newcols=list((colnames.values[Iloc==False]).astype(str))
        newcats=list((catnames[Iloc==False]).astype(str))

        # Make new columns in dataframe
        for col, cat in zip(newcols, newcats):
            simmat_padj[col]=np.nan
            Xlabx = np.append(Xlabx, cat)

        addrow=pd.DataFrame(index=newcols, columns=simmat_padj.columns.values).astype(float)
        simmat_padj=pd.concat([simmat_padj, addrow])
    return(simmat_padj, Xlabx)


# %% Remove columns without dtype
def _remove_columns_without_dtype(df, dtypes, verbose=3):
    if not isinstance(dtypes, str):
        assert df.shape[1]==len(dtypes), 'Columns in df and dtypes should match! [hnet.remove_columns_without_dtype]'
        Iloc = np.isin(dtypes, '')
        if np.any(Iloc):
            remcols=df.columns[Iloc].values
            df.drop(columns=remcols, inplace=True)
            dtypes=list(np.array(dtypes)[(Iloc==False)])
            if verbose>=3: print('[hnet] >%.0f columns are removed.' %(len(remcols)))

        assert df.shape[1]==len(dtypes), 'Columns in df and dtypes should match! [hnet.remove_columns_without_dtype]'

    return(df, dtypes)


# %% Clean empty rows
def _drop_empty(df, Xlabx, verbose=3):
    dfO=df.copy()
    cols=dfO.columns.values
    rows=dfO.index.values

    # Set diagonal on nan
    np.fill_diagonal(df.values, np.nan)

    droplabel=[]
    for col in cols:
        if np.any(cols==col):
            if np.all(np.logical_and(df.loc[:, cols==col].isna().values.reshape(-1, 1), df.loc[rows==col, :].isna().values.reshape(-1, 1))):
                if verbose>=3: print('[hnet] >Dropping %s' %(col))
                droplabel.append(col)

    # Remove labels from the original df
    Xlabx=Xlabx[np.isin(dfO.columns, droplabel)==False]
    dfO.drop(labels=droplabel, axis=0, inplace=True)
    dfO.drop(labels=droplabel, axis=1, inplace=True)

    return(dfO, Xlabx)


# %% Setup columns in correct dtypes
def _filter_significance(out, alpha, multtest):
    if isinstance(multtest, type(None)):
        idx=np.where(np.array(list(map(lambda x: x['P']<=alpha, out))))[0]
    else:
        idx=np.where(np.array(list(map(lambda x: x['Padj']<=alpha, out))))[0]

    outf = [out[i] for i in idx]
    if outf==[]: outf=None
    return(outf)


# %% Cleaning
def _nancleaning(datac, y):
    Iloc = datac.replace([np.inf, -np.inf, None, 'nan', 'None', 'inf', '-inf'], np.nan).notnull()
    datac = datac[Iloc]
    yc = y[Iloc]
    return(datac, yc)


# %% Do the math
def _post_processing(simmat_padj, nr_succes_pop_n, simmat_labx, alpha, multtest, fillna, dropna, verbose=3):
    # Clean label names by chaning X.0 into X
    simmat_padj.columns = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.columns))
    simmat_padj.index = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.index.values))
    nr_succes_pop_n = np.array(nr_succes_pop_n)
    nr_succes_pop_n[:, 0] = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, nr_succes_pop_n[:, 0]))

    if verbose>=5: print(simmat_padj)
    # Multiple test correction
    simmat_padj = _multipletestcorrectionAdjmat(simmat_padj, multtest, verbose=verbose)
    # Remove variables for which both rows and columns are empty
    if dropna: [simmat_padj, simmat_labx]=_drop_empty(simmat_padj, simmat_labx, verbose=verbose)
    # Fill empty fields
    if fillna: simmat_padj.fillna(1, inplace=True)
    # Remove those with P>alpha, to prevent unnecesarilly edges
    simmat_padj[simmat_padj>alpha]=1
    # Convert P-values to -log10 scale
    adjmatLog = _logscale(simmat_padj)

    # Set zeros on diagonal but make sure it is correctly ordered
    if np.all(adjmatLog.index.values==adjmatLog.columns.values):
        np.fill_diagonal(adjmatLog.values, 0)
    if np.all(simmat_padj.index.values==simmat_padj.columns.values):
        np.fill_diagonal(simmat_padj.values, 1)

    # Remove edges from matrix
    if dropna:
        idx1=np.where((simmat_padj==1).sum(axis=1)==simmat_padj.shape[0])[0]
        idx2=np.where((simmat_padj==1).sum(axis=0)==simmat_padj.shape[0])[0]
        keepidx= np.setdiff1d(np.arange(simmat_padj.shape[0]), np.intersect1d(idx1, idx2))
        simmat_padj=simmat_padj.iloc[keepidx, keepidx]
        adjmatLog=adjmatLog.iloc[keepidx, keepidx]
        simmat_labx=simmat_labx[keepidx]
        IA, _ = ismember(nr_succes_pop_n[:, 0], simmat_padj.columns.values)
        nr_succes_pop_n = nr_succes_pop_n[IA, :]

    return(simmat_padj, adjmatLog, simmat_labx, nr_succes_pop_n)


# %% Scale weights
def _scale_weights(weights, node_size_limits):
    out = MinMaxScaler(feature_range=(node_size_limits[0],node_size_limits[1])).fit_transform(np.append('0',weights).astype(float).reshape(-1,1)).flatten()[1:]
    return(out)


# %% Split filepath
def _path_split(filepath, rem_spaces=False):
    [dirpath, filename]=os.path.split(filepath)
    [filename,ext]=os.path.splitext(filename)
    if rem_spaces:
        filename=filename.replace(' ','_')
    return(dirpath, filename, ext)


# %% From savepath to full and correct path
def _path_correct(savepath, filename='fig', ext='.png'):
    """Correcth the path for filename.

    Description
    -----------
    savepath can be a string that looks like below.
    savepath='./tmp'
    savepath='./tmp/fig'
    savepath='./tmp/fig.png'
    savepath='fig.png'
    
    """
    out=None
    if not isinstance(savepath, type(None)):
        # Set savepath and filename
        [getdir, getfile, getext] = _path_split(savepath)
        # Make savepath
        if len(getdir[1:])==0: getdir=''
        if len(getfile)==0: getfile=filename
        if len(getext)==0: getext=ext
        # Create dir
        if len(getdir)>0:
            path=Path(getdir)
            path.mkdir(parents=True, exist_ok=True)
        # Make final path
        out=os.path.join(getdir,getfile + getext)
    return(out)


# %%
def _white_black_list(df, dtypes, white_list, black_list, verbose=3):
    # Keep only variables that are in white_list.
    if white_list is not None:
        white_list = [x.lower() for x in white_list]
        Iloc = np.isin(df.columns.str.lower(), white_list)
        df = df.loc[:,Iloc]
        if verbose>=3: print('[hnet] >Keeping ony features in the white list..')

    # Exclude variables that are in black_list.
    if black_list is not None:
        black_list = [x.lower() for x in black_list]
        Iloc = ~np.isin(df.columns.str.lower(), black_list)
        df = df.loc[:,Iloc]
        if verbose>=3: print('[hnet] >Removing features from the black list..')
        # Remove also in dtypes
        if not isinstance(dtypes, str):
            if len(dtypes)!=len(Iloc):
                raise ValueError('[hnet] >ERROR : dtypes should have same length as the input datafame! dtypes willbe automatically black listed too. Or do not specify dtypes.')
            else:
                dtypes = np.array(dtypes)[Iloc]

    # Check dtypes because things can have changed because of the white/black list
    if not isinstance(dtypes, str):
        if (white_list is not None) and len(dtypes)!=df.shape[1]:
            raise ValueError('[hnet] >ERROR : dtypes should have same length as the white_list! Or do not specify dtypes.')
        if len(dtypes)!=df.shape[1]:
            raise ValueError('[hnet] >ERROR : dtypes should have same length as your dataframe or white_list')

    if df.shape[1]<=1: print('[hnet] >Warning : After filtering, [%d] variable remained. A minimum of 2 is required. Tip: Check your dtypes, and see which ones are (not) categorical.' %(df.shape[1]))
    return df, dtypes


# %%
def _bool_processesing(df, dtypes, excl_background, verbose=3):
    if isinstance(dtypes, str) | (np.any(dtypes=='bool')):
        Iloc = ((df.dtypes=='bool') | (dtypes=='bool')).values
        if np.any(Iloc):
            if verbose>=3: print('[hnet] >Converting boolean values..')
            # Set as int
            df.loc[:, Iloc] = df.loc[:, Iloc].astype('int')
            # Remove the background values
            if (excl_background is None):
                excl_background = ['0.0']
            # Make list if required
            if isinstance(excl_background, str):
                excl_background = [excl_background]
            # Append to background
            excl_background = excl_background + ['0.0']
            # Get only unique background
            excl_background = np.unique(np.array(excl_background)).tolist()
            if verbose>=3: print('[hnet] >Set parameter <excl_background> to: %s.' %(str(excl_background)))
            if not isinstance(dtypes, str):
                # Set dtypes as catagorical
                dtypes = np.array(dtypes)
                dtypes[dtypes=='bool']='cat'
    # Return
    return df, dtypes, excl_background


# %% Preprocessing
def _preprocessing(df, dtypes='pandas', y_min=10, perc_min_num=0.8, excl_background=None, white_list=None, black_list=None, verbose=3):
    if verbose>=4: print('[hnet] >preprocessing: Column names are set to str. and spaces are trimmed.')
    df.columns = df.columns.astype(str)
    df.columns = df.columns.str.strip()
    df.reset_index(drop=True, inplace=True)

    # Convert bool columns to integer values
    df, dtypes, excl_background = _bool_processesing(df, dtypes, excl_background=excl_background, verbose=verbose)
    # Filter on white_list and black_list
    df, dtypes = _white_black_list(df, dtypes, white_list, black_list)
    # Remove columns without dtype
    df, dtypes = _remove_columns_without_dtype(df, dtypes, verbose=verbose)
    # Make onehot matrix for response variable y
    df_onehot = df2onehot.df2onehot(df, dtypes=dtypes, y_min=y_min, deep_extract=False, hot_only=True, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
    # Set the dtypes for the input dataframe
    df, dtypes = df2onehot.set_dtypes(df, dtypes=dtypes, deep_extract=False, perc_min_num=perc_min_num, num_if_decimal=True, verbose=0)
    dtypes = np.array(dtypes)

    # Make sure its limited to the number of y_min
    Iloc = (df_onehot['onehot'].sum(axis=0)>=y_min).values
    if np.any(Iloc==False):
        if verbose>=2: print('[hnet] >WARNING: Features with y_min needs another round of filtering. Fixing it now..')
        df_onehot['onehot']=df_onehot['onehot'].loc[:,Iloc]
        df_onehot['labx']=df_onehot['labx'][Iloc]

    # Some check before proceeding
    if (df_onehot['onehot'].empty) or (np.all(np.isin(dtypes, 'num'))): raise ValueError('[hnet] >ALL data is excluded from the dataframe! There should be at least 1 categorical value!')
    if df.shape[1] != len(dtypes): raise ValueError('[hnet] >DataFrame Shape and dtypes length does not match.')

    # Make all integer
    df_onehot['onehot'] = df_onehot['onehot'].astype(int)
    # Return
    return df, df_onehot, dtypes


# %% Tempdir
def _tempdir(savepath):
    if savepath is None:
        savepath = os.path.join(tempfile.gettempdir(), '')
    else:
        savepath = os.path.join(savepath, '')
    os.makedirs(savepath, exist_ok=True)

    return(savepath)


# %% Filter adjacency matrix
def _filter_adjmat(simmatLogP, labx, threshold=None, min_edges=None, white_list=None, black_list=None, verbose=3):
    # if (white_list is not None) and (len(white_list)<=1): raise ValueError('[hnet] >ERROR: white_list should contain at least 2 input variables.')

    # Filter on threshold
    if threshold is not None:
        if verbose>=3: print('[hnet] >Filtering associations on threshold > %d' %(threshold))
        simmatLogP[simmatLogP<threshold]=0

    # Filter on minimum number of edges
    if min_edges is not None:
        simmatBOOL = simmatLogP.copy()>0
        if np.all(simmatBOOL.columns==simmatBOOL.index.values):
            Iloc = np.logical_or(simmatBOOL.sum(axis=0)>=min_edges, simmatBOOL.sum(axis=1)>=min_edges).values
            if verbose>=3: print('[hnet] >Filtering on edges: [%d] variables remain after filtering on a minimum of [%d] edges.' %(np.sum(Iloc), min_edges))
            simmatLogP = simmatLogP.loc[Iloc,Iloc]
            labx = labx[Iloc]

    # Filter on white_list
    if white_list is not None:
        Irow = np.isin(simmatLogP.index.values, white_list)
        Icol = np.isin(simmatLogP.columns, white_list)
        Ilabx = np.isin(labx, white_list)
        Irow = np.logical_or(Irow, Ilabx)
        Icol = np.logical_or(Icol, Ilabx)
        if verbose>=3: print('[hnet] >Number of variables at input: [%d], and after white listing: [%d]' %(len(Irow), sum(Irow)))
        simmatLogP = simmatLogP.iloc[Irow,Icol]
        labx = labx[Irow]

    # Filter on white_list
    if black_list is not None:
        Irow = np.isin(simmatLogP.index.values, black_list)
        Icol = np.isin(simmatLogP.columns, black_list)
        Ilabx = np.isin(labx, black_list)
        Irow = np.logical_or(Irow, Ilabx)
        Icol = np.logical_or(Icol, Ilabx)
        if verbose>=3: print('[hnet] >Number of variables at input: [%d], and after black listing: [%d]' %(len(Irow), sum(~Irow)))
        simmatLogP = simmatLogP.iloc[~Irow, ~Icol]
        labx = labx[~Irow]

    return simmatLogP, labx



#%% Libraries
# pip install colour dash dash_bootstrap_components dash_core_components dash_html_components
import os
import base64
from pathlib import Path
# popular libraries
import numpy as np
import pandas as pd

# Front-end
# import Core
# import dash_dangerously_set_inner_html
# import codecs
import dash
# import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
import webbrowser
# Network center panel
import networkx as nx
import plotly.graph_objs as go
from colour import Color
# Hnet
import hnet as hnet
import pypickle
# progressbar
# import tkinter as tk
# from tkinter import ttk
# global labels
global TMP_DIRECTORY, HNET_DIR_STABLE, NETWORK_LAYOUT, HNET_OUT
# global edge1, node1

#%% Initializatoin
TMP_DIRECTORY     = './results/tmp/'
HNET_DIR_STABLE   = './results/stable/'
# BACKGROUND_IMAGE  = 'url(./static/background.jpg)'
TMP_DIRECTORY_DEL = False

# Create directories
path=Path(TMP_DIRECTORY)
path.mkdir(parents=True, exist_ok=True)
path=Path(HNET_DIR_STABLE)
path.mkdir(parents=True, exist_ok=True)

# At initialization remove content in the tmp directory
if TMP_DIRECTORY_DEL:
    print('[HNET-GUI] Cleaning files from tmp directory..')
    remfiles=os.listdir(TMP_DIRECTORY)
    for remfile in remfiles:
        if os.path.isfile(remfile): os.remove(os.path.join(TMP_DIRECTORY,remfile))

# Extract HNet results from tmp and stable directories
HNET_PATH_STABLE=[]
for i in os.listdir(HNET_DIR_STABLE):
    getdir=os.path.join(HNET_DIR_STABLE,i)
    if os.path.isdir(getdir):
        HNET_PATH_STABLE.append({'label':i,'value':getdir})

#%% NETWORK PROPERTIES
# ALPHA_SCORE=[0,1000]
ALPHA_SCORE=0
NODE_NAME=''
NETWORK_LAYOUT='fruchterman_reingold'
HNET_OUT=None
edge1=None
node1=None

#%%
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
#app.css.append_css({'external_url':external_stylesheets}) # Required for making columns
app.scripts.config.serve_locally = True
app.title = "HNet: Graphical Hypergeometric Networks"

#%% Start d3 network in browser
def boot_d3network_in_browser(dropdown_path, who=''):
    print('D3PATH: WHO: [%s]' %who)
    d3path=''
    if dropdown_path!=None:
        if os.path.isfile(get_d3path(dropdown_path)):
            print(d3path)
            d3path=os.path.abspath(get_d3path(dropdown_path))
            # open in webbrowser
            webbrowser.open(d3path, new=2)
            # https://plot.ly/python/network-graphs/
    return(d3path)

#%% Saving file
def save_file(name, content, savepath):
    # Decode and store file uploaded with plotly dash
    print('[HNET-GUI] Saving uploaded file..')
    data=content.encode('utf8').split(b";base64,")[1]
    filepath = os.path.join(savepath,name)
    with open(filepath, "wb") as fp:
        fp.write(base64.decodebytes(data))
    return(filepath)

#%% Check input params
def check_input(uploaded_filenames, uploaded_file_contents, y_min, alpha, k, excl_background, perc_min_num, specificity, multtest):
    runtxt=[]
    runOK=True
    dropna=[True]

    try:
        k=np.int(k)
    except:
        runtxt.append('k is a required parameter (k=1)\n')
        #k=1

    try:
        y_min=np.int(y_min)
    except:
        runtxt.append('y_min is a required parameter (y_min=10)\n')
        #y_min=10

    try:
        alpha=np.float(alpha)
    except:
        runtxt.append('alpha is a required parameter (alpha=0.05)\n')
        #alpha=0.05

    try:
        perc_min_num=np.float(perc_min_num)
    except:
        perc_min_num=None

    if specificity==None:
        runtxt.append('specificity is a required parameter (specificity=medium)\n')
        #specificity='medium'

    if len(dropna)>0:
        dropna=True
    else:
        dropna=False

    if excl_background=='':
        excl_background=None
    
    if (uploaded_filenames is None) or (uploaded_file_contents is None):
        runtxt.append('Input file is required\n')

    if len(runtxt)>0:
        runOK=False
    
    out=dict()
    out['k']=k
    out['dropna']=dropna
    out['y_min']=y_min
    out['alpha']=alpha
    out['multtest']=multtest
    out['perc_min_num']=perc_min_num
    out['specificity']=specificity
    out['excl_background']=excl_background
    out['uploaded_filenames']=uploaded_filenames
    out['uploaded_file_contents']=uploaded_file_contents
    return(out, runOK, runtxt)

#%% List result directories
def get_hnetpath(HNET_DIR_STABLE):
    HNET_PATH_STABLE=[]
    for i in os.listdir(HNET_DIR_STABLE):
        getdir=os.path.join(HNET_DIR_STABLE,i)
        if os.path.isdir(getdir):
            HNET_PATH_STABLE.append({'label':i,'value':getdir})
    return(HNET_PATH_STABLE)

# Make path for d3js
def get_d3path(filepath):
    return(os.path.join(filepath,'index.html'))

def get_pklpath(filepath):
    return(os.path.join(filepath,'hnet.pkl'))

#%% Create Network
# def network_graph_hnet(alphaRange, NodeToSearch, network_layout, HNET_OUT=None):
#     import plotly.graph_objects as go

#     if isinstance(HNET_OUT, type(None)):
#         edge_trace = go.Scatter(x=[0], y=[0], line=dict(width=0.5, color='#888'), hoverinfo='none', mode='lines')
#         node_trace = go.Scatter(x=[0], y=[0])
#     else:
#         G = HNET_OUT['G']['G']
#         # G = nx.random_geometric_graph(200, 0.125)
        
#         # Create edge traces
#         edge_x = []
#         edge_y = []
#         for edge in G.edges():
#             x0, y0 = HNET_OUT['G']['pos'][edge[0]]
#             x1, y1 = HNET_OUT['G']['pos'][edge[0]]
#             # x0, y0 = G.nodes[edge[0]]['pos']
#             # x1, y1 = G.nodes[edge[1]]['pos']
#             edge_x.append(x0)
#             edge_x.append(x1)
#             # edge_x.append(None)
#             edge_y.append(y0)
#             edge_y.append(y1)
#             # edge_y.append(None)
        
#         edge_trace = go.Scatter(x=edge_x, y=edge_y, line=dict(width=0.5, color='#888'), hoverinfo='none', mode='lines')
        
#         node_x = []
#         node_y = []
#         for node in G.nodes():
#             # x, y = G.nodes[node]['pos']
#             x, y = HNET_OUT['G']['pos'][node]
#             node_x.append(x)
#             node_y.append(y)
        
#         # Create node traces
#         node_trace = go.Scatter(
#             x=node_x, y=node_y,
#             mode='markers',
#             hoverinfo='text',
#             marker=dict(
#                 showscale=True,
#                 # colorscale options
#                 #'Greys' | 'YlGnBu' | 'Greens' | 'YlOrRd' | 'Bluered' | 'RdBu' |
#                 #'Reds' | 'Blues' | 'Picnic' | 'Rainbow' | 'Portland' | 'Jet' |
#                 #'Hot' | 'Blackbody' | 'Earth' | 'Electric' | 'Viridis' |
#                 colorscale='YlGnBu',
#                 reversescale=True,
#                 color=[],
#                 size=10,
#                 colorbar=dict(
#                     thickness=15,
#                     title='Node Connections',
#                     xanchor='left',
#                     titleside='right'
#                 ),
#                 line_width=2))
    
#         # Create node traces
#         node_adjacencies = []
#         node_text = []
#         for node, adjacencies in enumerate(G.adjacency()):
#             node_adjacencies.append(len(adjacencies[1]))
#             node_text.append('# of connections: '+str(len(adjacencies[1])))
        
#         node_trace.marker.color = node_adjacencies
#         node_trace.text = node_text
        
#     # Combine edge and node traces in figure
#     figure = go.Figure(data=[edge_trace, node_trace],
#                  layout=go.Layout(
#                     title='<br>Network graph made with Python',
#                     titlefont_size=16,
#                     showlegend=False,
#                     hovermode='closest',
#                     margin=dict(b=20,l=5,r=5,t=40),
#                     annotations=[ dict(
#                         text="HNET graph",
#                         showarrow=True,
#                         xref="paper", yref="paper",
#                         x=0.005, y=-0.002 ) ],
#                     xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
#                     yaxis=dict(showgrid=False, zeroline=False, showticklabels=False))
#                     )
#     # figure.show()
#     return(figure)

#%% Create Network
def network_graph(alphaRange, NodeToSearch, network_layout, HNET_OUT=None):
    # if (isinstance(edge1, type(None))) and (isinstance(node1, type(None))):
    #     print('demo case')
    if not os.path.isfile(os.path.join(TMP_DIRECTORY+'edge1.csv')):
        # Load demo dataset
        print('[HNET-GUI] DATA REQUIRED!')
    else:
       edge1 = pd.read_csv(os.path.join(TMP_DIRECTORY+'edge1.csv'))
       node1 = pd.read_csv(os.path.join(TMP_DIRECTORY+'node1.csv'))

    # filter the record by alpha score, to enable interactive control through the input box
    #print(alphaRange)
    I = edge1['Weight'].values >= np.min(alphaRange)
    edge1 = edge1.loc[I,:]
    edge1.reset_index(drop=True, inplace=True)
    nodeSet= np.unique(list(edge1['Source'])+list(edge1['Target']))

    # to define the centric point of the networkx layout
    shells=[]
    shell1=[]
    shell1.append(NodeToSearch)
    shells.append(shell1)
    shell2=[]
    for node in nodeSet:
        if node!=NodeToSearch:
            shell2.append(node)
    shells.append(shell2)
    
    # import network as network
    G = nx.from_pandas_edgelist(edge1, 'Source', 'Target', ['Source', 'Target', 'Weight'], create_using=nx.MultiDiGraph())
    nx.set_node_attributes(G, node1.set_index('NodeName')['Label'].to_dict(), 'Label')
    nx.set_node_attributes(G, node1.set_index('NodeName')['Type'].to_dict(), 'Type')

    # pos = nx.layout.spring_layout(G)
    # pos = nx.layout.circular_layout(G)
    # nx.layout.shell_layout only works for more than 3 nodes
    # print(len(NodeToSearch))

    if len(shell2)>1 and isinstance(NodeToSearch, type(None)) and NodeToSearch=='':
        pos = nx.layout.shell_layout(G, shells, scale=0.4, iterations=50)
    else:
        if network_layout=='fruchterman_reingold':
            pos = nx.layout.fruchterman_reingold_layout(G, scale=0.4, iterations=50)
        elif network_layout=='kamada_kawai':
            pos = nx.layout.kamada_kawai_layout(G, scale=0.4)
        elif network_layout=='circular':
            pos = nx.layout.circular_layout(G, scale=0.4)
        elif network_layout=='planar':
            pos = nx.layout.planar_layout(G, scale=0.4)
        else:
            pos = nx.layout.spring_layout(G, scale=0.4, iterations=50)

    for node in G.nodes:
        G.nodes[node]['pos'] = list(pos[node])


    # IF EMPTY
    if len(shell2)==0:
        print('Empty number of nodes selected.')
        traceRecode = []  # contains edge_trace, node_trace, middle_node_trace

        node_trace = go.Scatter(x=tuple([1]), y=tuple([1]), text=tuple([str(NodeToSearch)]), textposition="bottom center",
                                mode='markers+text',
                                marker={'size': 50, 'color': 'LightSkyBlue'})
        traceRecode.append(node_trace)

        node_trace1 = go.Scatter(x=tuple([1]), y=tuple([1]),
                                mode='markers',
                                marker={'size': 50, 'color': 'LightSkyBlue'},
                                opacity=0)
        traceRecode.append(node_trace1)

        figure = {
            "data": traceRecode,
            "layout": go.Layout(title='Network', showlegend=False,
                                margin={'b': 40, 'l': 40, 'r': 40, 't': 40},
                                xaxis={'showgrid': False, 'zeroline': False, 'showticklabels': False},
                                yaxis={'showgrid': False, 'zeroline': False, 'showticklabels': False},
                                height=600
                                )}
        return(figure)

    else:
        traceRecode = []  # contains edge_trace, node_trace, middle_node_trace
        #==============================================================================
        colors = list(Color('lightcoral').range_to(Color('darkred'), len(G.edges())))
        colors = ['rgb' + str(x.rgb) for x in colors]
    
        for index, edge in enumerate(G.edges):
            x0, y0 = G.nodes[edge[0]]['pos']
            x1, y1 = G.nodes[edge[1]]['pos']
            weight = float(G.edges[edge]['Weight']) / max(edge1['Weight']) * 10
            trace = go.Scatter(x=tuple([x0, x1, None]), y=tuple([y0, y1, None]),
                               mode='lines',
                               line={'width': weight},
                               marker=dict(color=colors[index]),
                               line_shape='spline',
                               opacity=1)
            traceRecode.append(trace)

        #==============================================================================
        node_trace = go.Scatter(x=[], y=[], hovertext=[], text=[], mode='markers+text', textposition="bottom center", hoverinfo="text", marker={'size': 50, 'color': 'LightSkyBlue'})
    
        for node in G.nodes():
            x, y = G.nodes[node]['pos']
            hovertext = "Label: " + str(G.nodes[node]['Label']) + "<br>" + "Label: " + str(G.nodes[node]['Type'])
            text = node#node1['NodeName'][index]
            node_trace['x'] += tuple([x])
            node_trace['y'] += tuple([y])
            node_trace['hovertext'] += tuple([hovertext])
            node_trace['text'] += tuple([text])
    
        traceRecode.append(node_trace)
        #==============================================================================
        middle_hover_trace = go.Scatter(x=[], y=[], hovertext=[], mode='markers', hoverinfo="text", marker={'size': 20, 'color': 'LightSkyBlue'}, opacity=0)
    
        for edge in G.edges:
            x0, y0 = G.nodes[edge[0]]['pos']
            x1, y1 = G.nodes[edge[1]]['pos']
            hovertext = "From: " + str(G.edges[edge]['Source']) + "<br>" + "To: " + str(
                G.edges[edge]['Target']) + "<br>" + str(
                G.edges[edge]['Weight']) + "<br>"
            middle_hover_trace['x'] += tuple([(x0 + x1) / 2])
            middle_hover_trace['y'] += tuple([(y0 + y1) / 2])
            middle_hover_trace['hovertext'] += tuple([hovertext])
    
        traceRecode.append(middle_hover_trace)
        #==============================================================================
        figure = {
            "data": traceRecode,
            "layout": go.Layout(title='Selected network', showlegend=False, hovermode='closest',
                                margin={'b': 40, 'l': 40, 'r': 40, 't': 40},
                                xaxis={'showgrid': False, 'zeroline': False, 'showticklabels': False},
                                yaxis={'showgrid': False, 'zeroline': False, 'showticklabels': False},
                                height=600,
                                clickmode='event+select',
                                annotations=[
                                    dict(
                                        ax=(G.nodes[edge[0]]['pos'][0] + G.nodes[edge[1]]['pos'][0]) / 2,
                                        ay=(G.nodes[edge[0]]['pos'][1] + G.nodes[edge[1]]['pos'][1]) / 2, axref='x', ayref='y',
                                        x=(G.nodes[edge[1]]['pos'][0] * 3 + G.nodes[edge[0]]['pos'][0]) / 4,
                                        y=(G.nodes[edge[1]]['pos'][1] * 3 + G.nodes[edge[0]]['pos'][1]) / 4, xref='x', yref='y',
                                        showarrow=True,
                                        arrowhead=3,
                                        arrowsize=4,
                                        arrowwidth=1,
                                        opacity=1
                                    ) for edge in G.edges]
                                )}
        return(figure)
    

#%% Setup webpage
GUIelements = html.Div([
        # Row 1
        html.Div([html.H5("HNets: Graphical Hypergeometric Networks")], className="row", style={'textAlign':'left','width':'100%','backgroundColor':'#e0e0e0'}),

        # Row 2 - Column 1
        html.Div([

            html.Div([
                html.H6('HNet input parameters'),
                dcc.Input(id='k-id', placeholder='Enter k..', type='text', value=1, style={"width": "100%"}), 
                #dcc.Input(id='alpha-id', placeholder='Enter alpha..', type='text', value=0.05, style={"width": "100%"}), 
                #dcc.Input(id='ymin-id', placeholder='Enter a value for y_min..', type='text', value=10, style={"width": "100%"}), 
                #dcc.Checklist(id='checkbox-id', options=[{'label':'drop nan','value':'True'}], value=['True'], style={"width": "100%"}),
                dcc.Input(id='perc_min_num-id', placeholder='Minimum percentage..', type='text', value='', style={"width": "100%"}), 
                dcc.Input(id='excl_background-id', placeholder='Remove background..', type='text', value='', style={"width": "100%"}), 
                dcc.Dropdown(id='alpha-id',
                    options=[
                        {'label': '0.0001', 'value': '0.0001'},
                        {'label': '0.001', 'value': '0.001'},
                        {'label': '0.01', 'value': '0.01'},
                        {'label': '0.05', 'value': '0.05'},
                        {'label': '0.1', 'value': '0.1'},
                        {'label': '1', 'value': '1'},
                    ],
                    value='0.05', style={"width": "100%"}),
                dcc.Dropdown(id='ymin-id',
                    options=[
                        {'label': '1', 'value': '1'},
                        {'label': '5', 'value': '5'},
                        {'label': '10', 'value': '10'},
                        {'label': '20', 'value': '20'},
                        {'label': '50', 'value': '50'},
                        {'label': '100', 'value': '100'},
                    ],
                    value='10', style={"width": "100%"}),
                dcc.Dropdown(id='specificity-id',
                    options=[
                        {'label': 'Low', 'value': 'low'},
                        {'label': 'Medium', 'value': 'medium'},
                        {'label': 'High', 'value': 'high'}
                    ],
                    value='medium', style={"width": "100%"}),
                dcc.Dropdown(id='multtest-id',
                    options=[
                        {'label': 'Holm', 'value': 'holm'},
                        {'label': 'Bonferroni', 'value': 'bonferroni'},
                        {'label': 'Hommel', 'value': 'hommel'},
                        {'label': 'Benjamini/Hochberg', 'value': 'fdr_bh'},
                        {'label': 'Benjamini/Yekutieli', 'value': 'fdr_by'},
                        {'label': 'Sidak', 'value': 'sidak'},
                        {'label': 'Holm-sidak', 'value': 'holm-sidak'},
                    ],
                    value='holm', style={"width": "100%"}),
                html.Div(id="message-box-output"),
                dcc.Upload(id="UPLOAD_BOX",children=html.Div(["Drag and drop or click to select a file to upload."]),
                       style={
                        #"width": "100%",
                        "height": "250px",
                        "lineHeight": "60px",
                        "borderWidth": "1px",
                        "borderStyle": "dashed",
                        "borderRadius": "5px",
                        "textAlign": "center",
                        "margin": "10px",
                        'backgroundColor':'white',
                    },
                    multiple=False), 

                # dcc.Dropdown(id='results-id', options=[{'label':i,'value':os.path.join(HNET_DIR_STABLE,i)} for i in os.listdir(HNET_DIR_STABLE)], value='', style={"width": "100%"}),
                dcc.Dropdown(id='opt-dropdown', options=get_hnetpath(HNET_DIR_STABLE), value='', style={"width": "100%"}),
                # dcc.Dropdown(id='opt-dropdown', value='', style={"width": "100%"}),
                # html.Div(id="results-output")

            ], className="three columns", style={"margin":"0px", "width": "15%", "border":"1px black solid", "height": "700px",'backgroundColor':''}),

            # ROWS 2 - COLUMN 2 --------------------- CENTER PANEL: NETWORK ------------------ 
            html.Div(className="three columns", children=[dcc.Graph(id="hnet-graph", figure=network_graph(ALPHA_SCORE, NODE_NAME, NETWORK_LAYOUT))], 
                     style={"margin":"0px","width": "65%", "height": "700px","border":"1px black solid"} ),

            # ROW 2 - COLUMN 3 -------------------- RIGHT PANEL: CONTROLS -------------------
            html.Div([html.H6('Network Settings'),

                dcc.Dropdown(id='network-layout-id',
                    options=[
                        {'label': 'kamada_kawai', 'value': 'kamada_kawai'},
                        {'label': 'planar', 'value': 'planar'},
                        {'label': 'circular', 'value': 'circular'},
                        {'label': 'shell', 'value': 'shell'},
                        {'label': 'spring', 'value': 'spring'},
                        {'label': 'fruchterman_reingold', 'value': 'fruchterman_reingold'},
                        {'label': 'graph', 'value': 'graph'}
                    ],
                    value='fruchterman_reingold', style={"width": "100%"}),

                dcc.Input(id='alpha-slider-id', placeholder='slider', type='text', value=0, style={"width": "100%"}), 
                html.Div(id="output-container-range-slider"),
                dcc.Input(id='node-id', placeholder='Node Name', type='text', style={"width": "100%"}),
                # html.Div(dcc.Input(id='message-box', type='text', style={"width": "100%"})),
                html.Button('Make d3graph', id='button', style={"width": "100%"}),
                html.Div(id='output-container-button',children='Enter a value and press submit', style={"width": "100%"}),
                # html.Button('Make d3graph!', id='button-id'),                
                # html.Div(id='output-button', children='Enter a value and press submit'),
                # html.Div(id="output"),
                ], className="three columns", style={"margin":"0px","width": "15%", "height": "700px","border":"1px black solid"}),
            

        ], className="row", style={"width": "100%"}),

        # html.Div(
        #     [dcc.Interval(id="progress-interval", n_intervals=0, interval=500),
        #      dbc.Progress(id="progress")],
        #     className="row", style={"width": "95%","border":"1px black solid"}
        #     ),


    ], className="row", style={"width": "100%"} #style={"max-width": "500px"},
)

#%% Run webpage
app.layout = html.Div([GUIelements])

#%%
# @app.callback(
#     [Output("progress", "value"), Output("progress", "children")],
#     [Input("progress-interval", "n_intervals")],
# )
# def update_progress(n):
#     https://dash-bootstrap-components.opensource.faculty.ai/l/components/progress
#     # check progress of some background process, in this example we'll just
#     # use n_intervals constrained to be in 0-100
#     progress = min(n % 110, 100)
#     # only add text after 5% progress to ensure text isn't squashed too much
#     return progress, f"{progress} %" if progress >= 5 else ""

#%% Callback for button (right panel)
@app.callback(
    Output('output-container-button', 'children'),
    [Input('button', 'n_clicks'), Input("opt-dropdown","value")])
def button_click_action(n_clicks, dropdown_path):
    if dropdown_path==None: dropdown_path=''
    d3path=boot_d3network_in_browser(dropdown_path, who='button')
    if d3path=='': d3path='Select first using the dropdown-box!'
    print(d3path)

#%% Callback for plotly (right-side) components
@app.callback(
    Output('hnet-graph', 'figure'),
    [Input('alpha-slider-id', 'value'), Input('node-id', 'value'), Input("opt-dropdown","value"), Input('network-layout-id', 'value')])
def update_output(alpha_limit, node_name, dropdown_path, network_layout):
    # Change input variables
    if alpha_limit=='': alpha_limit=0
    if alpha_limit==None: alpha_limit=0
    if network_layout=='': network_layout='fruchterman_reingold'
    if network_layout==None: network_layout='fruchterman_reingold'
    if dropdown_path==None: dropdown_path=''
    alpha_limit=np.int(alpha_limit)
    ALPHA_SCORE = np.int(alpha_limit)
    NODE_NAME = node_name
    NETWORK_LAYOUT = network_layout
    HNET_OUT = None
    # edge1=None
    # node1=None

    # Print selection dropdownbox
    # print(get_d3path(dropdown_path))
    # print(get_pklpath(dropdown_path))
    print('Dropdown_path path: %s\n' %(dropdown_path))
    print('Node name selected : %s\n' %(node_name))
    print('Node layout selected : %s\n' %(network_layout))
    # Load data
    if os.path.isfile(get_pklpath(dropdown_path)):
        print('Load picke file')
        HNET_OUT=pypickle.load(get_pklpath(dropdown_path))
        # HNET_OUT=hnet_out
        #print(df['simmatLogP'])

        df_edges=HNET_OUT['simmatLogP'].stack().reset_index()
        df_edges.columns=['Source', 'Target', 'Weight']
        df_edges['Weight']=df_edges['Weight'].astype(float)
        edge1 = df_edges.loc[df_edges['Weight']>0,:]
        node1 = pd.DataFrame(np.unique(df_edges[['Source','Target']].values.reshape(-1)), columns=['NodeName'])
        node1['Label']=node1['NodeName']
        node1['Type']=''
        
        # Write to disk
        edge1.to_csv(os.path.join(TMP_DIRECTORY+'edge1.csv'), index=False)
        node1.to_csv(os.path.join(TMP_DIRECTORY+'node1.csv'), index=False)
        # ALPHA_SCORE=[0,np.max(edge1['Weight'].values)]
        
    
    # Make graph
    # if (not isinstance(edge1, type(None))) and (not isinstance(node1, type(None))):
    hnet_graph=network_graph(alpha_limit, node_name, network_layout, HNET_OUT=HNET_OUT)
    # else:
    #     print('Select one first in dropdown!')
    # Return to screen
    return(hnet_graph)
    # to update the global variable of ALPHA_SCORE and NODE_NAME

#%% Callback for center screen
# @app.callback(
#     Output("results-output", "children"),
#     [Input("opt-dropdown","value")])
# def load_results(dropdown_path):
#     """Save uploaded files and regenerate the file list."""
#     # print(dropdown_path)
#     # d3path=boot_d3network_in_browser(dropdown_path)
#     # Return
#     return(dropdown_path)

#%% Callback for HNet menu (left-side)
@app.callback(
    Output("message-box-output", "children"),
    # Output('opt-dropdown', 'options'),
    [Input("UPLOAD_BOX","filename"), 
     Input("UPLOAD_BOX","contents"), 
     Input("ymin-id","value"), 
     Input("alpha-id","value"), 
     Input("k-id","value"),
     Input("excl_background-id","value"),
     Input("perc_min_num-id","value"),
     Input("specificity-id","value"),
     Input("multtest-id","value"),
     ],
)
def process_csv_file(uploaded_filenames, uploaded_file_contents, y_min, alpha, k, excl_background, perc_min_num, specificity, multtest):
    """Save uploaded files and regenerate the file list."""
    # Check input parameters
    [args, runOK, runtxt]=check_input(uploaded_filenames, uploaded_file_contents, y_min, alpha, k, excl_background, perc_min_num, specificity, multtest)
    if runOK==False:
        for txt in runtxt: print('[HNET-GUI] %s' %txt)
        return(runtxt)

    print('alpha:%s' %args['alpha'])
    print('y_min:%s' %args['y_min'])
    print('k:%s' %args['k'])
    print('multtest:%s' %args['multtest'])
    print('excl_background:%s' %args['excl_background'])
    print('perc_min_num:%s' %args['perc_min_num'])
    print('specificity:%s' %args['specificity'])
    print('dropna:%s' %args['dropna'])
    print('File input: %s' %(args['uploaded_filenames']))

#    if uploaded_filenames is not None and uploaded_file_contents is not None:
    filepath        = save_file(args['uploaded_filenames'], args['uploaded_file_contents'], TMP_DIRECTORY)
    [_,filename, _] = hnet.hnet.path_split(args['uploaded_filenames'])
    savepath        = os.path.join(HNET_DIR_STABLE,filename+'_'+str(args['y_min'])+'_'+str(args['k'])+'_'+str(args['multtest'])+'_'+str(args['specificity'])+'_'+str(args['perc_min_num'])+'_'+str(args['excl_background'])+'/')
    d3path          = get_d3path(savepath)
    pklpath         = get_pklpath(savepath)

    print('filepath %s' %(filepath))
    print('savepath %s' %(savepath))
    print('d3path %s' %(d3path))
    print('pklpath %s' %(pklpath))

    # Progressbar
    # mGUI=tk.Tk()
    # mGUI.title('HNET '+ filename)
    # progressbar = ttk.Progressbar(mGUI, orient='horizontal', length=300)
    # progressbar.pack(side=tk.TOP)
    # progressbar.config(mode='determinate')
    # progressbar.update()
    # progressbar['value']=0
    # progressbar['maximum']=5
    # # Set starting position
    # progressbar['value']=progressbar['value']+1 # update bar
    # progressbar.update() # update gui

    # Make directory
    if not os.path.isdir(savepath):
        os.mkdir(savepath)
    # Make D3js path
    if not os.path.isfile(d3path):
        # Read file
        df = pd.read_csv(filepath)
        # Run HNet
        # progressbar['value']=progressbar['value']+1 # update bar
        # progressbar.update() # update gui
        HNET_OUT = hnet.fit(df, alpha=args['alpha'], y_min=args['y_min'], k=args['k'], multtest=args['multtest'], dtypes='pandas', specificity=args['specificity'], perc_min_num=args['perc_min_num'], dropna=args['dropna'], excl_background=args['excl_background'], verbose=3)
        # Save pickle file
        print('SAVING NETWORK FIGURE: %s' %(savepath))
        # progressbar['value']=progressbar['value']+1 # update bar
        # progressbar.update() # update gui
        HNET_OUT['G'] = hnet.plot_network(HNET_OUT, dist_between_nodes=0.4, scale=2, dpi=250, figsize=[30,20], showfig=False, savepath=os.path.join(savepath,'hnet_network.png'))
        # Store pickle file
        print('STORE PICKLE')
        # progressbar['value']=progressbar['value']+1 # update bar
        # progressbar.update() # update gui
        pypickle.save(pklpath, HNET_OUT)
        #print('MAKE D3GRAPH')
        _ = hnet.plot_d3graph(HNET_OUT, savepath=savepath, showfig=False)
    else:
        print('dir exists, load stuff')
        HNET_OUT=pypickle.load(pklpath)
        
    # Open in browser
    if os.path.isfile(d3path):
        print('OPEN BROWSER')
        webbrowser.open(os.path.abspath(d3path), new=2)

    print('HNET_DIR_STABLE: %s!' %(HNET_DIR_STABLE))
    print('%s done!' %(filename))
    print('-----------------------Done!-----------------------')
    # progressbar['value']=progressbar['maximum'] # update bar
    # progressbar.update() # update gui
    # try:
    #     mGUI.destroy()
    # except:
    #     pass
    return(filename)


#%% Main
if __name__ == "__main__":
    # webbrowser.open('http://127.0.0.1:8050/', new=0, autoraise=True)
    app.run_server(debug=True, use_reloader=True, port=8050)



from hnet.hnet import hnet

from hnet.hnet import (
    enrichment,
    compare_networks,
    to_undirected,
    import_example,
    )

from hnet.utils.adjmat_vec import (
    vec2adjmat,
    adjmat2vec,
    )

__author__ = 'Erdogan Tasksen'
__email__ = 'erdogant@gmail.com'
__version__ = '1.1.5'

# module level doc-string
__doc__ = """
HNET - Graphical Hypergeometric networks.
=====================================================================

Description
-----------
Creation of networks from datasets with mixed datatypes and with unknown function.
Input datasets can range from generic dataframes to nested data structures with lists, missing values and enumerations.
HNet (graphical Hypergeometric Networks) is a method where associations across variables are tested for significance by statistical inference.
The aim is to determine a network with significant associations that can shed light on the complex relationships across variables.

Examples
--------
>>> # Initialize hnet with default settings
>>> from hnet import hnet
>>> hn = hnet()
>>> # Load example dataset
>>> df = hn.import_example('titanic')

>>> # Structure learning
>>> out = hn.association_learning(df)

>>> # Plot dynamic graphs
>>> G = hn.d3graph()
>>> G = hn.d3heatmap()
>>> # Plot static graph
>>> G = hn.plot()
>>> # Plot heatmap
>>> hn.heatmap(cluster=True)
>>> # Plot feature importance
>>> hn.plot_feat_importance()

>>> # Examine differences between models
>>> import hnet
>>> scores, adjmat = hnet.compare_networks(out['simmatP'], out['simmatP'], showfig=True)
>>> adjmat_undirected = hnet.to_undirected(out['simmatLogP'])

References
----------
* https://erdogant.github.io/hnet/
* https://github.com/erdogant/hnet
* https://arxiv.org/abs/2005.04679

"""



""" This function has functionalities for network creation, clustering and plotting.

	G    = network.to_graph(adjmat)
	labx = network.cluster(G)
	fig  = network.plot(G, labx)

 INPUT:
   adjmat:         pd.DataFrame(): Edges are any values >0 between intersecting nodes.
                   rows    = features
                   colums  = samples
 OPTIONAL

   verbose:        Integer [0..5] if verbose >= DEBUG: print('debug message')
                   0: (default)
                   1: ERROR
                   2: WARN
                   3: INFO
                   4: DEBUG
                   

 OUTPUT
	output

 DESCRIPTION
   This function has functionalities for network creation, clustering and plotting
 
 REQUIRES
   pip install python-louvain
   pip install --upgrade networkx
   

 EXAMPLE
   %reset -f
   %matplotlib auto
   import matplotlib.pyplot as plt
   import numpy as np
   import networkx as nx
   import NETWORKS.network as network
   
   adjmat=np.random.randint(0,2,(10,10))

   G = nx.karate_club_graph()
   adjmat = nx.adjacency_matrix(G).todense()
   
   
   # Create network
   G = network.to_graph(adjmat)
   
   # Cluster
   [G,labx]=network.cluster(G)
   [*G.node.values()]
   
   # Plot
   fig=network.plot(G, node_color=labx, node_size=500)
   
   # Plot interactive
   from NETWORKS.d3graph import d3graph
   import pandas as pd
   G = nx.karate_club_graph()
   nodes=pd.DataFrame([*G.node.values()])['club']
   adjmat = nx.adjacency_matrix(G).todense()
   adjmat=pd.DataFrame(index=nodes, data=adjmat, columns=nodes)

   A=d3graph(adjmat, node_color=labx.astype(str))
   A=d3graph(adjmat)
   

 SEE ALSO
    d3graph
"""

# --------------------------------------------------------------------------
# Name        : network.py
# Version     : 1.0
# Author      : E.Taskesen
# Contact     : erdogant@gmail.com
# Date        : July. 2019
# --------------------------------------------------------------------------

# %% Libraries
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from community import community_louvain
import networkx as nx
from sklearn.preprocessing import minmax_scale
from sklearn.preprocessing import MinMaxScaler 
import df2onehot
from ismember import ismember
import classeval as clf

#%% Make graph from adjacency matrix
def to_graph(adjmat, verbose=3):
    assert float(nx.__version__)>2, 'This function requires networkx to be v2 or higher. Try to: pip install --upgrade networkx'
    config = dict()
    config['verbose'] = verbose

    adjmat = df2onehot.is_DataFrame(adjmat)
    if config['verbose']>=3: print('[hnet] >Making graph..')
    G=nx.from_pandas_adjacency(adjmat)

    return(G)

#%% Convert Adjmat to graph (G) (also works with lower versions of networkx)
def adjmat2graph(adjmat):
    G = nx.DiGraph() # Directed graph
    # Convert adjmat to source target
    df_edges=adjmat.stack().reset_index()
    df_edges.columns=['source', 'target', 'weight']
    df_edges['weight']=df_edges['weight'].astype(float)
    
    # Add directed edge with weigth
    for i in range(df_edges.shape[0]):
        if df_edges['weight'].iloc[i]!=0:
            # Setup color
            if df_edges['weight'].iloc[i]==1:
                color='k'
            elif df_edges['weight'].iloc[i]>1:
                color='r'
            elif df_edges['weight'].iloc[i]<0:
                color='b'
            else:
                color='p'
            
            # Create edge in graph
            G.add_edge(df_edges['source'].iloc[i], df_edges['target'].iloc[i], weight=np.abs(df_edges['weight'].iloc[i]), color=color)    
    # Return
    return(G)
    
#%% Compute similarity matrix
def compute_centrality(G, centrality='betweenness', verbose=3):
    if verbose>=3: print('[hnet] >Computing centrality %s' %(centrality))
    
    if centrality=='betweenness':
        bb=nx.centrality.betweenness_centrality(G)
    elif centrality=='closeness':
        bb=nx.centrality.closeness_centrality(G)
    elif centrality=='eigenvector':
        bb=nx.centrality.eigenvector_centrality(G)
    elif centrality=='degree':
        bb=nx.centrality.degree_centrality(G)
    elif centrality=='edge':
        bb=nx.centrality.edge_betweenness(G)
    elif centrality=='harmonic':
        bb=nx.centrality.harmonic_centrality(G)
    elif centrality=='katz':
        bb=nx.centrality.katz_centrality(G)
    elif centrality=='local':
        bb=nx.centrality.local_reaching_centrality(G)
    elif centrality=='out_degree':
        bb=nx.centrality.out_degree_centrality(G)
    elif centrality=='percolation':
        bb=nx.centrality.percolation_centrality(G)
    elif centrality=='second_order':
        bb=nx.centrality.second_order_centrality(G)
    elif centrality=='subgraph':
        bb=nx.centrality.subgraph_centrality(G)
    elif centrality=='subgraph_exp':
        bb=nx.centrality.subgraph_centrality_exp(G)
    elif centrality=='information':
        bb=nx.centrality.information_centrality(G)
    else:
        print('[NETWORK] [ERROR] Centrality <%s> does not exist!' %(centrality))
    
    # Set the attributes
    score=np.array([*bb.values()])
    nx.set_node_attributes(G, bb, centrality)
    
    return(G, score)

#%% compute clusters
def cluster(G, verbose=3):
    if verbose>=3: print('[hnet] >Clustering using best partition')
    # Partition
    partition=community_louvain.best_partition(G)
    # Set property to node
    nx.set_node_attributes(G, partition, 'clusterlabel')
    # Extract labels
    labx=[partition.get(node) for node in G.nodes()]
    labx=np.array(labx)
    
    return(G, labx)

#%% Compute cluster comparison
def cluster_comparison_centralities(G, width=5, height=4, showfig=False, methodtype='default', verbose=3):
    config=dict()
    config['showfig']=showfig
    config['width']=width
    config['height']=height
    config['verbose']=verbose

    if verbose>=3: print('[NETWORK] Compute a dozen of centralities and clusterlabels')
    
    # compute labx for each of the centralities
    centralities=['betweenness', 'closeness','eigenvector','degree','edge','harmonic','katz','local','out_degree','percolation','second_order','subgraph','subgraph_exp','information']
    
    # Compute best positions for the network
    pos=nx.spring_layout(G)

    # Cluster data nd store label in G
    [G, score] = cluster(G)
    
    # Compute centrality score for each of the centralities and store in G
    for centrality in centralities:
        [G,score]=compute_centrality(G, centrality=centrality)
    
    # Store
    df=pd.DataFrame([*G.nodes.values()])
    df.set_index(np.array([*G.nodes.keys()]), inplace=True)
    
    # Make plots
    for centrality in centralities:
        if config['showfig']:
            plot(G, node_color=df['clusterlabel'], node_size=df[centrality], pos=pos, cmap='Set1', title=centrality, width=config['width'], height=config['height'], methodtype='default')
    
    return(G, df)

#%% Make plot
def plot(G, node_color=None, node_label=None, node_size=100, node_size_scale=[25,200], alpha=0.8, font_size=18, cmap='Set1', width=40, height=30, pos=None, filename=None, title=None, methodtype='default', verbose=3):
    # https://networkx.github.io/documentation/networkx-1.7/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html
    config=dict()
    config['filename']=filename
    config['width']=width
    config['height']=height
    config['verbose']=verbose
    config['node_size_scale']=node_size_scale
    
    if verbose>=3: print('[NETWORK] Creating network plot')
    
    if 'pandas' in str(type(node_size)):
        node_size=node_size.values
    
    #scaling node sizes
    if config['node_size_scale']!=None and 'numpy' in str(type(node_size)):
        if verbose>=3: print('[NETWORK] Scaling node sizes')
        node_size=minmax_scale(node_size, feature_range=(node_size_scale[0], node_size_scale[1]))

    # Node positions
#    if isinstance(pos, type(None)):
#        pos=nx.spring_layout(G)

#    if isinstance(node_label, type(None)):
#        node_label=[*G.nodes])

    fig=plt.figure(figsize=(config['width'], config['height']))
    
    # Make the graph
    if methodtype=='circular':
        nx.draw_circular(G, labels=node_label, node_size=node_size, alhpa=alpha, node_color=node_color, cmap=cmap, font_size=font_size, with_labels=True)
    elif methodtype=='kawai':
        nx.draw_kamada_kawai(G, labels=node_label, node_size=node_size, alhpa=alpha, node_color=node_color, cmap=cmap, font_size=font_size, with_labels=True)
    else:
        nx.draw_networkx(G, labels=node_label, pos=pos, node_size=node_size, alhpa=alpha, node_color=node_color, cmap=cmap, font_size=font_size, with_labels=True)
#        nx.draw_networkx(G, pos=pos, node_size=node_size, alhpa=alpha, node_color=node_color, cmap=cmap, font_size=font_size)
        
    plt.title(title)
    plt.grid(True)
    plt.show()
    
    # Savefig
    if not isinstance(config['filename'], type(None)):
        if verbose>=3: print('[hnet] >Saving figure..')
        plt.savefig(config['filename'])
    
    return(fig)

#%% Normalize in good d3 range
def normalize_size(getsizes, minscale=0.1, maxscale=4):
    getsizes = MinMaxScaler(feature_range=(minscale,maxscale)).fit_transform(getsizes).flatten()
    return(getsizes)

#%% Convert dataframe to Graph
def df2G(df_nodes, df_edges, verbose=3):
    # Put edge information in G
#    G = nx.from_pandas_edgelist(df_edges, 'source', 'target', ['weight', 'edge_weight','edge_width','source_label','target_label'])
    
    colnames=list(df_edges.columns.values[~np.isin(df_edges.columns.values,['source','target'])])
    G = nx.from_pandas_edgelist(df_edges, 'source', 'target', colnames)

    # Put node info in G
    getnodes=[*G.nodes]
    for col in df_nodes.columns:
        for i in range(0, df_nodes.shape[0]):
            if np.any(np.isin(getnodes, df_nodes.index.values[i])):
                G.nodes[df_nodes.index.values[i]][col] = str(df_nodes[col].iloc[i])
    
    return(G)
    
#%% Convert Graph to dataframe
def G2df(G, node_color=None, node_label=None, node_size=100, edge_distance_minmax=[1,100], verbose=3):
    # Nodes
    df_node_names=pd.DataFrame([*G.nodes], columns=['node_name'])
    df_node_props=pd.DataFrame([*G.nodes.values()])
    df_nodes=pd.concat([df_node_names, df_node_props], axis=1)

    if not np.any(df_nodes.columns=='node_color'):
        df_nodes['node_color']='#000080'
    if not np.any(df_nodes.columns=='node_color_edge'):
        df_nodes['node_color_edge']='#000000'
    if not np.any(df_nodes.columns=='node_size_edge'):
        df_nodes['node_size_edge']=1
    if not isinstance(node_label, type(None)):
        df_nodes['node_label']=node_label
    
    if np.any(df_nodes.columns=='node_size'):
        df_nodes['node_size']=normalize_size(df_nodes['node_size'].values.reshape(-1,1), 1, 10)
    else:
        df_nodes['node_size']=10

    # Edges
    df_edge_links=pd.DataFrame([*G.edges], columns=['source_label','target_label'])
    df_edge_props=pd.DataFrame([*G.edges.values()])
    df_edges=pd.concat([df_edge_links, df_edge_props], axis=1)
    
    # Source and target values
    df_nodes['index_value']=None
    df_edges['source']=None
    df_edges['target']=None
#    uinodes=np.unique(np.append(df_edges['source_label'], df_edges['target_label']))
    uinodes=np.unique(df_nodes['node_name'])
    for i in range(0,len(uinodes)):
        I=(uinodes[i]==df_edges['source_label'])
        df_edges['source'].loc[I]=i
        I=(uinodes[i]==df_edges['target_label'])
        df_edges['target'].loc[I]=i
        
        I=df_nodes['node_name']==uinodes[i]
        df_nodes['index_value'].loc[I]=i
    
    df_nodes.set_index(df_nodes['index_value'], inplace=True)
    del df_nodes['index_value']
    
    # Include width and weights
    if not np.any(df_edges.columns=='edge_weight') and np.any(df_edges.columns=='weight'):
        df_edges['edge_weight']=normalize_size(df_edges['weight'].values.reshape(-1,1), edge_distance_minmax[0], edge_distance_minmax[1])
    else:
        df_edges['edge_weight']=2
    if not np.any(df_edges.columns=='edge_width'):
        df_edges['edge_width']=2
    if not np.any(df_edges.columns=='weight'):
        df_edges['weight']=2

    # Remove self-loops
    I=df_edges['source']!=df_edges['target']
    df_edges=df_edges.loc[I,:].reset_index(drop=True)

    return(df_nodes, df_edges)

#%% Make plot
def bokeh(G, node_color=None, node_label=None, node_size=100, node_size_scale=[25,200], alpha=0.8, font_size=18, cmap='Set1', width=40, height=30, pos=None, filename=None, title=None, methodtype='default', verbose=3):
    import networkx as nx
    from bokeh.io import show, output_file
    from bokeh.models import Plot, Range1d, MultiLine, Circle, HoverTool, BoxZoomTool, ResetTool
    from bokeh.models.graphs import from_networkx
    from bokeh.palettes import Spectral4
    
    SAME_CLUB_COLOR, DIFFERENT_CLUB_COLOR = "black", "red"
    edge_attrs = {}
    
    for start_node, end_node, _ in G.edges(data=True):
        edge_color = SAME_CLUB_COLOR if G.nodes[start_node]["club"] == G.nodes[end_node]["club"] else DIFFERENT_CLUB_COLOR
        edge_attrs[(start_node, end_node)] = edge_color
    
    nx.set_edge_attributes(G, edge_attrs, "edge_color")
    
    # Show with Bokeh
    plot = Plot(plot_width=400, plot_height=400,
                x_range=Range1d(-1.1, 1.1), y_range=Range1d(-1.1, 1.1))
    plot.title.text = "Graph Interaction Demonstration"
    
    node_hover_tool = HoverTool(tooltips=[("index", "@index"), ("club", "@club")])
    plot.add_tools(node_hover_tool, BoxZoomTool(), ResetTool())
    
    graph_renderer = from_networkx(G, nx.spring_layout, scale=1, center=(0, 0))
    
    graph_renderer.node_renderer.glyph = Circle(size=15, fill_color=Spectral4[0])
    graph_renderer.edge_renderer.glyph = MultiLine(line_color="edge_color", line_alpha=0.8, line_width=1)
    plot.renderers.append(graph_renderer)
    
    output_file("interactive_graphs.html")
    show(plot)


# %% Comparison of two networks
def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15, height=8, verbose=3):
    """Compare two networks.

    Parameters
    ----------
    adjmat_true : TYPE
        Adjacency matrix 1.
    adjmat_pred : TYPE
        Adjacency matrix 2.
    pos : list, optional
        list with coordinates to orientate the nodes.
    showfig : bool, optional
        Plot figure to screen. The default is True.
    width : int, optional
        Width of the figure. The default is 15.
    height : int, optional
        Height of the figure. The default is 8.
    verbose : int, optional
        Verbosity. The default is 3.

    Returns
    -------
    dict.
        scores
        adjmat_diff
    """
    # Make sure columns and indices to match
    IArow, IBrow = ismember(adjmat_true.index.values, adjmat_pred.index.values)
    IAcol, IBcol = ismember(adjmat_true.columns.values, adjmat_pred.columns.values)
    adjmat_true = adjmat_true.loc[IArow, IAcol]
    adjmat_pred = adjmat_pred.iloc[IBrow, IBcol]

    # Check whether order is correct
    if not np.all(adjmat_true.columns.values==adjmat_pred.columns.values): raise ValueError('Column order of both input values could not be matched')
    if not np.all(adjmat_true.index.values==adjmat_pred.index.values): raise ValueError('Row order of both input values could not be matched')

    # Make sure it is boolean adjmat
    adjmat_true = adjmat_true>0
    adjmat_pred = adjmat_pred>0

    # Get edges
    y_true = adjmat_true.stack().reset_index()[0].values
    y_pred = adjmat_pred.stack().reset_index()[0].values

    # Compute score
    scores = clf.confmatrix.eval(y_true, y_pred, verbose=verbose)
    # bayes.plot(out_bayes['adjmat'], pos=G['pos'])

    # Setup graph
    # adjmat_diff = adjmat_true - adjmat_pred
    adjmat_diff = adjmat_true.astype(int)
    adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))<0]=2
    adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))>0]=-1

    if showfig:
        # Setup graph
        # G_true = adjmat2graph(adjmat_true)
        G_diff = adjmat2graph(adjmat_diff)
        # Graph layout
        pos = graphlayout(G_diff, pos=pos, scale=1, layout='fruchterman_reingold')
        # Bootup figure
        plt.figure(figsize=(width,height))
        # nodes
        nx.draw_networkx_nodes(G_diff, pos, node_size=700, with_labels=True)
        # edges
        colors  = [G_diff[u][v]['color'] for u,v in G_diff.edges()]
        #weights = [G_diff[u][v]['weight'] for u,v in G_diff.edges()]
        nx.draw_networkx_edges(G_diff, pos, arrowstyle='->', edge_color=colors, width=1)
        # Labels
        nx.draw_networkx_labels(G_diff, pos, font_size=20, font_family='sans-serif')
        # Get labels of weights
        #labels = nx.get_edge_attributes(G,'weight')
        # Plot weights
        nx.draw_networkx_edge_labels(G_diff, pos, edge_labels=nx.get_edge_attributes(G_diff,'weight'))
        # Making figure nice
        #plt.legend(['Nodes','TN','FP','test'])
        ax = plt.gca()
        ax.set_axis_off()
        plt.show()
    
    # Return
    return(scores, adjmat_diff)

#%% Make graph layout
def graphlayout(model, pos, scale=1, layout='fruchterman_reingold', verbose=3):
    if isinstance(pos, type(None)):
        if layout=='fruchterman_reingold':
            pos = nx.fruchterman_reingold_layout(model, scale=scale, iterations=50)
        else:
            pos = nx.spring_layout(model, scale=scale, iterations=50)
    else:
        if verbose>=3: print('[hnet] >Existing coordinates from <pos> are used for the graphlayout.')

    return(pos)


import hnet
print(dir(hnet))
print(hnet.__version__)

# %%
import numpy as np
import pandas as pd

# %% Titanic case
from hnet import hnet
hn = hnet()
df = hn.import_example('titanic')
del df['PassengerId']
del df['Name']
results = hn.association_learning(df)

# STATIC
hn.heatmap(summarize=True, cluster=False)
hn.plot(node_color='cluster')

# Feature importance
hn.plot_feat_importance(marker_size=50)

# %%
from hnet import hnet
hn = hnet()
df = hn.import_example('sprinkler')

hn1 = hnet(excl_background='0.0', dtypes=np.array(['bool']*df.shape[1]))
out1 = hn1.association_learning(df.astype(bool))

hn2 = hnet()
out2 = hn2.association_learning(df.astype(bool))

hn2.d3graph(summarize=True)
hn2.d3graph()

hn2.d3heatmap(summarize=True)
hn2.d3heatmap(summarize=False)

hn2.heatmap(summarize=True, cluster=False)
hn2.heatmap(cluster=True)

# %%
import hnet
df = hnet.import_example('sprinkler')
out = hnet.enrichment(df.astype(bool), y=df.iloc[:, 0].values)
print(out)

import hnet
dtypes = np.array(['bool']*df.shape[1])
df = hnet.import_example('sprinkler')
out = hnet.enrichment(df.astype(bool), y=df.iloc[:, 0].values, dtypes=dtypes)
print(out)

# %%
# Load data
# df = hn.import_example('titanic')
from hnet import hnet
hn = hnet()
df = hn.import_example('sprinkler')

hn = hnet()
out2 = hn.association_learning(df.astype(bool))


hn = hnet(excl_background='0.0')
out1 = hn.association_learning(df)
out3 = hn.association_learning(df.astype(int))
hn.plot()

hn = hnet(excl_background='False')
hn.association_learning(df.astype(bool).astype(str))
hn.plot(summarize=True)

hn = hnet(excl_background='0')
out7 = hn.association_learning(df.astype(int).astype(str))
hn.plot(summarize=False)
hn.plot(summarize=True)

hn = hnet(excl_background='0.0')
out6 = hn.association_learning(df.astype(float).astype(str))
hn.plot(summarize=False)
hn.plot(summarize=True)

# Should raise exception
out4 = hn.association_learning(df.astype(float))
hn.plot(summarize=False)
hn.plot(summarize=True)


# %% Import class
from hnet import hnet
print(dir(hnet))

# %% Download dataset from url
# from tabulate import tabulate
import hnet
df = hnet.import_example(url='https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data')
df.columns=['age','workclass','fnlwgt','education','education-num','marital-status','occupation','relationship','race','sex','capital-gain','capital-loss','hours-per-week','native-country','earnings']
cols_as_float = ['age','hours-per-week','capital-loss','capital-gain']
df[cols_as_float]=df[cols_as_float].astype(float)

# print(tabulate(df.head(), tablefmt="grid", headers="keys"))

from hnet import hnet
hn = hnet(black_list=['fnlwgt'])
results = hn.association_learning(df)

# out = tabulate(hn.results['rules'].iloc[1:,:].head(), tablefmt="grid", headers="keys")

hn.d3heatmap(figsize=(1000,1000))
hn.d3graph(node_color='cluster')
hn.d3graph(min_edges=5)
# hn.d3graph(savepath='D://PY/REPOSITORIES/erdogant.github.io/docs/d3graph/income/')


# %% Import examples
import hnet

df = hnet.import_example('titanic')

df = hnet.import_example('student')

df = hnet.import_example('sprinkler')

# %% Run with default settings
import hnet
df = hnet.import_example('titanic')
from hnet import hnet
hn = hnet()
results = hn.association_learning(df)

# %% Run with default settings
from hnet import hnet

hn = hnet()
# Load data
# df = hn.import_example('titanic')
df = hn.import_example('sprinkler')
# Association learning
results = hn.association_learning(df.astype(bool))

# %% Plot with clustering nodes
G_static = hn.plot()
G_dynamic = hn.d3graph()

# Plot dynamic graph
G_dynamic = hn.d3graph(node_color='cluster')

# Plot static graph
G_static = hn.plot(node_color='cluster')

# Plot heatmap
P_heatmap = hn.heatmap(cluster=True)

# Plot heatmap
paths = hn.d3heatmap()

# %%
import matplotlib.pyplot as plt
P = hn.results['simmatLogP'].values.ravel()
P = P[P>0]

plt.figure(figsize=(15,10))
plt.hist(P, bins=50)
plt.grid(True)
plt.xlabel('-log10(P)')
plt.ylabel('Frequency')
plt.title('P-value Associaties')


# %% Association learning
hn1 = hnet(y_min=50)
hn2 = hnet()
# Data
df = hn1.import_example('titanic')
# Association learning
out1 = hn1.association_learning(df)
out2 = hn2.association_learning(df)


# Import hnet functionalities
import hnet
# Examine differences between models
[scores, adjmat] = hnet.compare_networks(out1['simmatLogP'], out2['simmatLogP'], showfig=True)

# Make undirected matrix
adjmat_undirected = hnet.to_undirected(out['simmatLogP'])


# %% Enrichment
import hnet

df = hnet.import_example('titanic')
y = df['Survived'].values
out = hnet.enrichment(df, y)

# %% White list and black list examples
from hnet import hnet
dtypes=['', 'cat', 'cat', '', 'cat', 'num', 'cat', 'cat', 'cat', 'num', 'cat', 'cat']
hn = hnet(black_list=['Survived','Embarked','Fare','CABIN'], dtypes=dtypes)
df = hn.import_example('titanic')
out = hn.association_learning(df)

hn = hnet(black_list=['Survived','Embarked','Fare','CABIN'])
dtypes=['', 'cat', 'cat', '', 'cat', 'num', 'cat', 'cat', 'cat', 'num', 'cat', 'cat']
df = hn.import_example('titanic')
out = hn.association_learning(df)

hn = hnet(white_list=['Survived','Embarked','Fare','Cabin'], dtypes=['num','cat','cat','cat'])
df = hn.import_example('titanic')
out = hn.association_learning(df)

hn = hnet(white_list=['Survived','Embarked','Fare','Cabin'])
df = hn.import_example('titanic')
out = hn.association_learning(df)
# hn.d3graph()

hn = hnet(black_list=['Survived','Embarked','Fare','CABIN'])
df = hn.import_example('titanic')
out = hn.association_learning(df)


# %% Run with blacklist
from hnet import hnet
hn = hnet(black_list=['PassengerId', 'Survived', 'Pclass', 'Name', 'Sex', 'Age', 'SibSp'])
# Load data
df = hn.import_example('titanic')
# Association learning
out = hn.association_learning(df)

# %% Run with whitelist
from hnet import hnet
hn = hnet(white_list=['Survived', 'Pclass', 'Age', 'SibSp'])
# Load data
df = hn.import_example('titanic')
# Association learning
out = hn.association_learning(df)

# %% Sklearn examples
import pandas as pd
from hnet import hnet

from sklearn.datasets import fetch_kddcup99
df = fetch_kddcup99(return_X_y=False)
df = pd.DataFrame(data=df['data'])
df.columns = df.columns.astype(str)
hn = hnet()
out = hn.association_learning(df)


# %%
from hnet import hnet
import pandas as pd
from tabulate import tabulate


df = hn.import_example('cancer')
hn = hnet(black_list=['tsneX','tsneY','PC1','PC2'])
results = hn.association_learning(df)
hn.d3graph(black_list=['sex'])
out = tabulate(hn.results['rules'].iloc[1:,:].head(), tablefmt="grid", headers="keys")
print(tabulate(df.head(), tablefmt="grid", headers="keys"))

# %% Small retail
df = hn.import_example('retail')
hn1 = hnet()
results1 = hn1.association_learning(df)
hn1.d3graph(savepath='D://PY/REPOSITORIES/erdogant.github.io/docs/d3graph/marketing_data_online_retail_small/')
# hn1.plot()
# hn1.d3graph()
# hn1.heatmap()

# %%

df = hn.import_example('waterpump')
hn2 = hnet(black_list=['id','longitude','latitude'])
results2 = hn2.association_learning(df)
hn2.d3graph(savepath='D://PY/REPOSITORIES/erdogant.github.io/docs/d3graph/waterpump/')
# hn2.plot()
# hn2.d3graph()
# hn2.heatmap()

# df['id'].head().values

# %%
df = hn.import_example('fifa')


hn3 = hnet(dtypes=['None', 'cat', 'cat', 'cat', 'num',
       'num', 'num', 'num', 'num', 'num', 'num',
       'num', 'num', 'num', 'num',
       'num', 'cat', 'cat',
       'cat', 'cat', 'cat', 'cat', 'cat', 'cat',
       'cat', 'cat', 'num'])

results3 = hn3.association_learning(df)
hn3.d3graph()
# hn3.d3graph(savepath='D://PY/REPOSITORIES/erdogant.github.io/docs/d3graph/fifa_2018/')
# hn3.plot()
# hn3.heatmap()

out = tabulate(hn3.results['rules'].iloc[1:,:].head(), tablefmt="grid", headers="keys")
print(tabulate(df.head(), tablefmt="grid", headers="keys"))

# hn3.plot()
# hn3.plot(directed=False, black_list=['Man of the Match_No'], node_color='cluster')
# hn3.plot(directed=True, black_list=['Man of the Match_No'], node_color='cluster')
# hn3.plot(directed=True, black_list=['Man of the Match_No'])
# hn3.d3graph(directed=False, black_list=['Man of the Match_No'])
# hn3.heatmap(black_list=['Man of the Match_No'])

# %% Download dataset from url
from tabulate import tabulate
import hnet
df = hnet.import_example(url='https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data')
df.columns=['age','workclass','fnlwgt','education','education-num','marital-status','occupation','relationship','race','sex','capital-gain','capital-loss','hours-per-week','native-country','earnings']
cols_as_float = ['age','hours-per-week','capital-loss','capital-gain']
df[cols_as_float]=df[cols_as_float].astype(float)

# print(tabulate(df.head(), tablefmt="grid", headers="keys"))

from hnet import hnet
hn = hnet(black_list=['fnlwgt'])
results = hn.association_learning(df)

out = tabulate(hn.results['rules'].iloc[1:,:].head(), tablefmt="grid", headers="keys")

hn.d3graph(savepath='D://PY/REPOSITORIES/erdogant.github.io/docs/d3graph/income/')
hn.d3graph(min_edges=1)

# %% Covid-19 dataset
from hnet import hnet
import pandas as pd
df = pd.read_csv('D://covid19_us_county.zip')

dtypes = ['', 'cat', 'cat', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num', 'num']
hn = hnet(dtypes=dtypes)
results = hn.association_learning(df)

# %% Police shooting
# https://github.com/washingtonpost/data-police-shootings
import pandas as pd
from hnet import hnet

# Load results
results = hn.load(filepath='C://temp/New folder/data-police-shootings/hnet.pkl')

# Run hnet
if results is None:
    # Read csv
    df = pd.read_csv('C://temp/New folder/data-police-shootings/fatal-police-shootings-data.csv')
    df['month'] = pd.to_datetime(df['date']).dt.month_name()
    df.head()
    # Initialize
    hn = hnet(excl_background=['nan'], black_list=['id', 'date', 'name'])
    # Association learning
    results = hn.association_learning(df, verbose=4)
    # Save
    hn.save(filepath='C://temp/New folder/data-police-shootings/hnet.pkl', overwrite=True)

# Plot dynamic graph
hn.d3graph()
hn.d3graph(black_list=['city'])
hn.d3graph(black_list=['city', 'state'])
hn.d3graph(summarize=True)

# Plot heatmap
hn.d3heatmap(summarize=True, vmax=2)
hn.d3heatmap(black_list=['city', 'state', 'longitude', 'latitude'])

# Feature importance
hn.plot_feat_importance()



"""HNET: Graphical Hypergeometric-networks."""
# -------------------------------------------------
# Name        : hnet.py
# Author      : E.Taskesen
# Contact     : erdogant@gmail.com
# github      : https://github.com/erdogant/hnet
# Licence     : See licences
# -------------------------------------------------

# %% Libraries
import warnings
warnings.filterwarnings("ignore")

# Custom packages
from d3graph import d3graph as d3graphs
from d3heatmap import d3heatmap as d3
from ismember import ismember
import imagesc
import df2onehot
import colourmap

# Local utils
import pypickle
from hnet.utils.savefig import savefig
import hnet.utils.network as network
import hnet.utils.hnstats as hnstats

# Known libraries
from scipy.stats import combine_pvalues
from sklearn.preprocessing import MinMaxScaler, LabelEncoder
label_encoder = LabelEncoder()

import networkx as nx

# Internal
import wget
import os
import pandas as pd
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt

# from functools import lru_cache

# %% Association learning across all variables
class hnet():
    """HNET - Graphical Hypergeometric networks.

    Description
    -----------
    This is the main function to detect significant edge probabilities between pairs of vertices (node-links) given the input DataFrame.

    A multi-step process is performed which consisting 5 steps.
        1. Pre-processing: Typing and One-hot Enconding. Each feature is set as being categoric, numeric or is excluded. The typing can be user-defined or automatically determined on conditions. Encoding of features in a one-hot dense array is done for the categoric terms. The one-hot dense array is subsequently used to create combinatory features using k combinations over n features (without replacement).
        2. Combinations: Make smart combinations between features because many mutual exclusive classes do exists.
        3. Hypergeometric test: The final dense array is used to assess significance with the categoric features.
        4. Wilcoxon Ranksum: To assess significance across the numeric features (Xnumeric) in relation to the dense array (Xcombination), the Mann-Whitney-U test is performed.
        5. Multiple test correction: Declaring significance for node-links.

    The final output of HNet is an adjacency matrix containing edge weights that depicts the strength of pairs of vertices.
    The adjacency matrix can then be examined as a network representation using d3graph.

    Parameters
    ----------
    alpha : float [0..1], (default : 0.05)
        Significance to keep only edges with <=alhpa.
        1 : (for all results)
    y_min : int [1..n], where n is the number of samples. (default : 10)
        Minimum number of samples in a group. Should be [samples>=y_min]. All groups with less then y_min samples are labeled as _other_ and are not used in the model.
        10, None, 1, etc
    k : int, [1..n] , (default : 1)
        Number of combinatoric elements to create for the n features
    perc_min_num : float, (default : 0.8)
        Force column (int or float) to be numerical if unique non-zero values are above percentage.
        Can also be None.
    multtest : String, (default : 'holm')
        * None: No multiple Test,
        * 'bonferroni': one-step correction,
        * 'sidak': one-step correction,
        * 'holm-sidak': step down method using Sidak adjustments,
        * 'holm': step-down method using Bonferroni adjustments,
        * 'simes-hochberg': step-up method  (independent),
        * 'hommel': closed method based on Simes tests (non-negative),
        * 'fdr_bh': Benjamini/Hochberg  (non-negative),
        * 'fdr_by': Benjamini/Yekutieli (negative),
        * 'fdr_tsbh': two stage fdr correction (non-negative),
        * 'fdr_tsbky': two stage fdr correction (non-negative)
    dtypes : list of str, (default : 'pandas')
        list strings, example: ['cat','num',''] of length y. By default the dtype is determined based on the pandas dataframe. Empty ones [''] are skipped.
        Can also be of the form: ['cat','cat','num','','cat']
    specificity : String, (default : 'medium')
        Configure how numerical data labels are stored. Setting this variable can be of use in the 'association_learning' function for the creation of a network ([None] will glue most numerical labels together whereas [high] mostly will not).
        * None : No additional information in the labels,
        * 'low' : 'high' or 'low' are included that represents significantly higher or lower assocations compared to the rest-group,
        * 'medium': 'high' or 'low' are included with 1 decimal behind the comma,
        * 'high' : 'high' or 'low' are included with 3 decimal behind the comma.
    dropna : Bool, [True,False] (Default : True)
        Drop rows/columns in adjacency matrix that showed no significance
    excl_background : list or None, [0], [0, '0.0', 'male', ...],  (Default: None)
        Remove values/strings that labeled as background. As an example, in a two-class approach with [0,1], the 0 is usually the background and not of interest.
        Example: '0.0': To remove categorical values with label 0
    black_list : List or None (default : None)
        If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
    white_list : List or None (default : None)
        If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.

    Returns
    -------
    dict()
        simmatP : pd.DataFrame()
            Adjacency matrix containing P-values between variable assocations.
        simmatLogP :  pd.DataFrame()
            -log10(P-value) of the simmatP.
        labx :  list of str
            Labels that are analyzed.
        dtypes :  list of str
            dtypes that are set for the labels.
        counts :  list of str
            Relative counts for the labels based on the number of successes in population.

    Examples
    --------
    >>> from hnet import hnet
    >>> hn = hnet()
    >>> # Load example dataset
    >>> df = hn.import_example('sprinkler')
    >>> association Learning
    >>> out = hn.association_learning(df)
    >>> # Plot dynamic graph
    >>> G_dynamic = hn.d3graph()
    >>> # Plot static graph
    >>> G_static = hn.plot()
    >>> # Plot heatmap
    >>> P_heatmap = hn.heatmap(cluster=True)
    >>> # Plot feature importance
    >>> hn.plot_feat_importance()
    """

    def __init__(self, alpha=0.05, y_min=10, perc_min_num=0.8, k=1, multtest='holm', dtypes='pandas', specificity='medium', dropna=True, excl_background=None, black_list=None, white_list=None):
        """Initialize distfit with user-defined parameters."""
        if (alpha is None): alpha=1
        if (y_min is None): y_min=1
        if isinstance(white_list, str): white_list=[white_list]
        if isinstance(black_list, str): black_list=[black_list]
        if (white_list is not None) and len(white_list)==0: white_list=None
        if (black_list is not None) and len(black_list)==0: black_list=None
        # Store in object
        self.alpha = alpha
        self.y_min = np.maximum(1, y_min)
        self.k = k
        self.multtest = multtest
        self.dtypes = dtypes
        self.specificity = specificity
        self.perc_min_num = perc_min_num
        self.dropna = dropna
        self.fillna = True
        self.excl_background = excl_background
        self.white_list = white_list
        self.black_list = black_list

    def prepocessing(self, df, verbose=3):
        """Pre-processing based on the model parameters."""
        # Pre processing
        [df, df_onehot, dtypes] = hnstats._preprocessing(df, dtypes=self.dtypes, y_min=self.y_min, perc_min_num=self.perc_min_num, excl_background=self.excl_background, white_list=self.white_list, black_list=self.black_list, verbose=verbose)
        # Add combinations
        [X_comb, X_labx, X_labo] = hnstats._make_n_combinations(df_onehot['onehot'], df_onehot['labx'], self.k, self.y_min, verbose=verbose)
        # Get numerical columns
        colNum = df.columns[df.dtypes == 'float64'].values
        simmat_labx = np.append(X_labo, colNum).astype(str)
        simmatP = pd.DataFrame(index=np.append(X_comb.columns, colNum).astype(str), columns=np.append(X_comb.columns, colNum).astype(str)).astype(float)
        # Return
        return df, simmatP, simmat_labx, X_comb, X_labx, dtypes

    def compute_associations(self, df, simmatP, simmat_labx, X_comb, X_labx, dtypes, verbose=3):
        """Association learning on the processed data."""
        # Here we go! in parallel!
        # from multiprocessing import Pool
        # nr_succes_pop_n=[]
        # with Pool(processes=os.cpu_count()-1) as pool:
        #     for i in range(0,X_comb.shape[1]):
        #         result = pool.apply_async(_do_the_math, (df, X_comb, dtypes, X_labx, param, i,))
        #         nr_succes_pop_n.append(result)
        #     results = [result.get() for result in result_objs]
        #     print(len(results))

        # Print some
        if verbose>=3: print('[hnet] >Association learning across [%d] categories.' %(X_comb.shape[1]))

        disable = (True if (verbose==0 or verbose>3) else False)
        nr_succes_pop_n = []

        for i in tqdm(range(0, X_comb.shape[1]), disable=disable):
            nr_succes_i, simmatP, simmat_labx = _do_the_math(df, X_comb, dtypes, X_labx, simmatP, simmat_labx, i, self.specificity, self.y_min, verbose=verbose)
            nr_succes_pop_n.append(nr_succes_i)
            if verbose>=5: print('[hnet] >[%d] %s' %(i, nr_succes_i))

        # Post processing
        simmatP, simmatLogP, simmat_labx, nr_succes_pop_n = hnstats._post_processing(simmatP, nr_succes_pop_n, simmat_labx, self.alpha, self.multtest, self.fillna, self.dropna, verbose=verbose)
        # Message
        print_text = '[hnet] >Total number of associatons computed: [%.0d]' %((simmatP.shape[0] * simmatP.shape[1]) - simmatP.shape[0])
        if verbose>=3: print('-' * len(print_text))
        if verbose>=3: print(print_text)
        if verbose>=3: print('-' * len(print_text))
        # Return
        return simmatP, simmatLogP, simmat_labx, nr_succes_pop_n

    def association_learning(self, df, verbose=3):
        """Learn the associations in the data.

        Parameters
        ----------
        df : DataFrame, [NxM].
            N=rows->samples, and  M=columns->features.

           |    | f1| f2| f3|
           | s1 | 0 | 0 | 1 |
           | s2 | 0 | 1 | 0 |
           | s3 | 1 | 1 | 0 |

        verbose : int [1-5], default: 3
            Print information to screen. A higher number will print more.

        Returns
        -------
        dict.
        simmatP : pd.DataFrame()
            Adjacency matrix containing P-values between variable assocations.
        simmatLogP :  pd.DataFrame()
            -log10(P-value) of the simmatP.
        labx :  list of str
            Labels that are analyzed.
        dtypes :  list of str
            dtypes that are set for the labels.
        counts :  list of str
            Relative counts for the labels based on the number of successes in population.

        """
        if not isinstance(df, pd.DataFrame): raise ValueError('Input data [df] must be of type pd.DataFrame()')
        # Preprocessing
        df, simmatP, labx, X_comb, X_labx, dtypes = self.prepocessing(df, verbose=verbose)
        # Here we go! Over all columns now
        simmatP, simmatLogP, labx, nr_succes_pop_n = self.compute_associations(df, simmatP, labx, X_comb, X_labx, dtypes, verbose=verbose)
        # Combine rules
        rules = self.combined_rules(simmatP, labx, verbose=0)
        # Feature importance
        feat_importance = self._feat_importance(simmatLogP, labx, verbose=verbose)
        # Compute associations for the catagories
        GsimmatP, GsimmatLogP = self._compute_associations_cat(simmatP, labx, verbose=verbose)
        # Store
        self.results = _store(simmatP, simmatLogP, GsimmatP, GsimmatLogP, labx, df, nr_succes_pop_n, dtypes, rules, feat_importance)
        # Use this option for storage of your model
        if verbose>=3: print('[hnet] >Fin.')
        return self.results

    def _compute_associations_cat(self, simmatP, labx, verbose=3):
        """Compute category associations by fishers method."""
        # Make some checks
        if not np.all(simmatP.columns==simmatP.index.values):
            raise ValueError('[hnet] >Error: Adjacency matrix [simmatP] does not have the same number of columns and index!')
        if not np.all(simmatP.shape[0]==len(labx)):
            raise ValueError('[hnet] >Error: The number of columns in [simmatP] does not match with the label [labx]!')

        if verbose>=3: print('[hnet] >Computing category association using fishers method..')
        uilabx = np.unique(labx)
        adjmatP = np.ones((len(uilabx), len(uilabx)))

        # Combine Pvalues for catagories based on fishers method
        for labxC in tqdm(uilabx, disable=(True if verbose==0 else False)):
            for labxR in uilabx:
                Ic = labx==labxC
                Ir = labx==labxR
                adjmatP[uilabx==labxR, uilabx==labxC] = combine_pvalues(simmatP.loc[Ir, Ic].values.ravel(), method='fisher')[1]

        # Store in dataframe
        adjmatP = pd.DataFrame(data=adjmatP, index=uilabx, columns=uilabx)
        adjmatlogP = -np.log10(adjmatP)
        adjmatlogP[adjmatlogP<=0]=0
        return adjmatP, adjmatlogP

    def _feat_importance(self, simmatLogP, labx, verbose=3):
        """Compute feature importance."""
        # Get unique labels
        uilabx, counts = np.unique(labx, return_counts=True)
        # Count the number of significant associations per category
        Pcounts = []
        Psum = []
        for lab in uilabx:
            Iloc = labx==lab
            # Sum over significance per label
            Psum.append(simmatLogP.iloc[Iloc, :].sum(axis=1).sum())
            # Count number of significant category labels
            Pcounts.append((simmatLogP.iloc[Iloc, :]>0).sum(axis=1).sum())

        # Compute corrected Pvalue
        Pcorr = (np.array(Psum) / np.array(Pcounts))
        Pcorr[np.isnan(Pcorr)]=0
        # Store
        df_feat_importance = pd.DataFrame(data=np.c_[Pcounts, Psum, Pcorr], columns=['degree', 'Psum', 'Pcorr'])
        df_feat_importance.index = uilabx
        df_feat_importance.sort_values(by='Pcorr', ascending=False, inplace=True)
        # Return
        return df_feat_importance

    # Make network d3
    def plot_feat_importance(self, marker_size=5, top_n=10, figsize=(15, 8), verbose=3):
        """Plot feature importance.

        Parameters
        ----------
        marker_size : int, (default: 5)
            Marker size in the scatter plot.
        top_n : int, optional
            Top n features are labelled in the plot. The default is 10.
        figsize : tuple, optional
            Size of the figure in the browser, [height,width]. The default is [1500,1500].
        verbose : int, optional
            Verbosity. The default is 3.

        Returns
        -------
        None.

        """
        # Make plot counts significant associations per label
        df = pd.DataFrame((self.results['simmatLogP']>0).sum(axis=1), columns=['Psum'])
        df.reset_index(drop=False, inplace=True)
        c = colourmap.fromlist(self.results['labx'])[0]
        plt.figure(figsize=figsize)
        plt.scatter(np.arange(0, len(df.values)), df['Psum'].values, color=c, s=marker_size)
        plt.grid(True)
        plt.xlabel('label')
        plt.ylabel('Degree')
        plt.title('Number of significant associations per label (degree)')
        # Add the top n labels
        top_n = np.minimum(top_n, self.results['feat_importance'].shape[0])
        df = df.sort_values(by='Psum', ascending=False)[0:top_n]
        for i in range(0, top_n):
            plt.text(df.index[i], df['Psum'].iloc[i], df['index'].iloc[i])

        # Plot the cataogires
        plt.figure(figsize=figsize)
        self.results['feat_importance']['degree'].plot(kind='bar', figsize=(15, 8))
        plt.grid(True)
        plt.xlabel('Catagories')
        plt.ylabel('Total degree (significant associations)')
        plt.title('Total degree per category; the number of significant edges.')

        # Plot feature importance after normalization
        plt.figure(figsize=figsize)
        self.results['feat_importance']['Pcorr'].plot(kind='bar')
        plt.grid(True)
        plt.xlabel('Catagories')
        plt.ylabel('Normalized significance')
        plt.title('Feature importance after normalization. Per category: E(-log10(P) / E(significant edges per node)')

    # Get adjacency matrix
    def _get_adjmat(self, summarize):
        """Retrieve data for catagories or labels."""
        if summarize:
            simmatP = self.results['simmatP_cat']
            simmatLogP = self.results['simmatLogP_cat']
            labx = self.results['simmatP_cat'].columns.values
        else:
            simmatP = self.results['simmatP']
            simmatLogP = self.results['simmatLogP']
            labx = self.results['labx']
        return simmatP, simmatLogP, labx

    # Make network d3
    def d3heatmap(self, summarize=False, savepath=None, directed=True, threshold=None, white_list=None, black_list=None, min_edges=None, figsize=(700, 700), vmax=None, showfig=True, verbose=3):
        """Interactive heatmap creator.

        Description
        -----------
        This function creates a interactive and stand-alone heatmap that is build on d3 javascript.
        d3heatmap is integrated into hnet and uses the -log10(P-value) adjacency matrix.
        Each column and index name represents a node whereas values >0 in the matrix represents an edge.
        Node links are build from rows to columns. Building the edges from row to columns only matters in directed cases.
        The network nodes and edges are adjusted in weight based on hte -log10(P-value), and colors are based on the category names.

        Parameters
        ----------
        self : Object
            The output of .association_learning()
        summarize : bool, (default: False)
            Show the results based on categoric or label-specific associations.
            True: Summrize based on the categories
            False: All associations across labels
        savepath : str
            Save the figure in specified path.
        directed : bool, default is True.
            Create network using directed edges (arrows).
        threshold : int (default : None)
            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
        black_list : List or None (default : None)
            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
        white_list : List or None (default : None)
            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
        min_edges : int (default : None)
            Edges are only shown if a node has at least min_edges.
        showfig : bool, optional
            Plot figure to screen. The default is True.
        figsize : tuple, optional
            Size of the figure in the browser, [height,width]. The default is [1500,1500].

        Returns
        -------
        dict : containing various results derived from network.
        savepath : str
            Save the figure in specified path.
        labx : array-like
            Cluster labels.

        """
        if verbose>=3: print('[hnet] >Building dynamic heatmaps using d3graph..')
        # Check results
        status = self._check_results()
        if not status: return None
        # Retrieve data
        _, simmatLogP, labx = self._get_adjmat(summarize)

        # Filter adjacency matrix on blacklist/whitelist and/or threshold
        simmatLogP, labx = hnstats._filter_adjmat(simmatLogP, labx, threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
        # Check whether anything has remained
        if simmatLogP.values.flatten().sum()==0:
            if verbose>=3: print('[hnet] >Nothing to plot.')
            return None

        # Make undirected network
        if not directed:
            simmatLogP = to_undirected(simmatLogP, method='logp')

        # Cluster
        labx = self.plot(summarize=summarize, node_color='cluster', directed=True, threshold=threshold, white_list=white_list, black_list=black_list, min_edges=min_edges, showfig=False)['labx']

        if vmax is None:
            vmax = np.max(np.max(simmatLogP)) / 10

        # Make heatmap
        if verbose>=3: print('[hnet] >Creating output html..')
        paths = d3.heatmap(simmatLogP, clust=labx, path=savepath, title='Hnet d3heatmap', vmax=vmax, width=figsize[1], height=figsize[0], showfig=showfig, stroke='red', verbose=verbose)

        # Return
        results = {}
        results['paths'] = paths
        results['clust_labx'] = labx
        return(results)

    # Make network d3
    def d3graph(self, summarize=False, node_size_limits=[6, 15], savepath=None, node_color=None, directed=True, threshold=None, white_list=None, black_list=None, min_edges=None, charge=500, figsize=(1500, 1500), showfig=True, verbose=3):
        """Interactive network creator.

        Description
        -----------
        This function creates a interactive and stand-alone network that is build on d3 javascript.
        d3graph is integrated into hnet and uses the -log10(P-value) adjacency matrix.
        Each column and index name represents a node whereas values >0 in the matrix represents an edge.
        Node links are build from rows to columns. Building the edges from row to columns only matters in directed cases.
        The network nodes and edges are adjusted in weight based on hte -log10(P-value), and colors are based on the category names.

        Parameters
        ----------
        self : Object
            The output of .association_learning()
        summarize : bool, (default: False)
            Show the results based on categoric or label-specific associations.
            True: Summrize based on the categories
            False: All associations across labels
        node_size_limits : tuple
            node sizes are scaled between [min,max] values. The default is [6,15].
        savepath : str
            Save the figure in specified path.
        node_color : None or 'cluster' default : None
            color nodes based on clustering or by label colors.
        directed : bool, default is True.
            Create network using directed edges (arrows).
        threshold : int (default : None)
            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
        black_list : List or None (default : None)
            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
        white_list : List or None (default : None)
            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
        min_edges : int (default : None)
            Edges are only shown if a node has at least min_edges.
        showfig : bool, optional
            Plot figure to screen. The default is True.
        figsize : tuple, optional
            Size of the figure in the browser, [height,width]. The default is [1500,1500].

        Returns
        -------
        dict : containing various results derived from network.
        G : graph
            Graph generated by networkx.
        savepath : str
            Save the figure in specified path.
        labx : array-like
            Cluster labels.

        """
        if verbose>=3: print('[hnet] >Building dynamic network graph using d3graph..')
        # Check input data
        status = self._check_results()
        if not status: return None
        # Retrieve data
        _, simmatLogP, labx = self._get_adjmat(summarize)
        # Setup tempdir
        savepath = hnstats._tempdir(savepath)
        # Filter adjacency matrix on blacklist/whitelist and/or threshold
        simmatLogP, labx = hnstats._filter_adjmat(simmatLogP, labx, threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
        # Check whether anything has remained
        if simmatLogP.values.flatten().sum()==0:
            if verbose>=3: print('[hnet] >Nothing to plot.')
            return None

        # Resizing nodes based on user-limits
        IA, IB = ismember(simmatLogP.columns, self.results['counts'][:, 0])
        node_size = np.repeat(node_size_limits[0], len(simmatLogP.columns))
        node_size[IA] = hnstats._scale_weights(self.results['counts'][IB, 1], node_size_limits)

        # Make undirected network
        if not directed:
            simmatLogP = to_undirected(simmatLogP, method='logp')

        # Color node using network-clustering
        if node_color=='cluster':
            labx = self.plot(summarize=summarize, node_color='cluster', directed=True, threshold=threshold, white_list=white_list, black_list=black_list, min_edges=min_edges, showfig=False)['labx']
        else:
            labx = label_encoder.fit_transform(labx)

        # Make network
        if verbose>=3: print('[hnet] >Creating output html..')
        Gout = d3graphs(simmatLogP.T, savepath=savepath, node_size=node_size, charge=charge, height=figsize[0], width=figsize[1], collision=0.1, node_color=labx, directed=directed, showfig=showfig)
        # Return
        Gout['labx'] = labx
        return(Gout)

    # Make network plot
    def plot(self, summarize=False, scale=2, dist_between_nodes=0.4, node_size_limits=[25, 500], directed=True, node_color=None, savepath=None, figsize=[15, 10], pos=None, layout='fruchterman_reingold', dpi=250, threshold=None, white_list=None, black_list=None, min_edges=None, showfig=True, verbose=3):
        """Make plot static network plot of the model results.

        Description
        -----------
        The results of hnet can be vizualized in several manners, one of them is a static network plot.

        Parameters
        ----------
        self : Object
            The output of .association_learning()
        summarize : bool, (default: False)
            Show the results based on categoric or label-specific associations.
            True: Summrize based on the categories
            False: All associations across labels
        scale : int, optional
            scale the network by blowing it up by scale. The default is 2.
        dist_between_nodes : float, optional
            Distance between the nodes. Edges are sized based this value. The default is 0.4.
        node_size_limits : int, optional
            Nodes are scaled between the Min and max size. The default is [25,500].
        node_color : str, None or 'cluster' default is None
            color nodes based on clustering or by label colors.
        directed : bool, default is True.
            Create network using directed edges (arrows).
        savepath : str, optional
            Save the figure in specified path.
        figsize : tuple, optional
            Size of the figure, [height,width]. The default is [15,10].
        pos : list, optional
            list with coordinates to orientate the nodes.
        layout : str, optional
            layouts from networkx can be used. The default is 'fruchterman_reingold'.
        dpi : int, optional
            resolution of the figure. The default is 250.
        threshold : int (default : None)
            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
        black_list : List or None (default : None)
            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
        white_list : List or None (default : None)
            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
        min_edges : int (default : None)
            Edges are only shown if a node has at least min_edges.
        showfig : bool, optional
            Plot figure to screen. The default is True.

        Returns
        -------
        dict. Dictionary containing various results derived from network. The keys in the dict contain the following results:
        G : graph
            Graph generated by networkx.
        labx : str
            labels of the nodes.
        pos :  list
            Coordinates of the node postions.

        """
        if verbose>=3: print('[hnet] >Building static network graph..')
        # Check input data
        status = self._check_results()
        if not status: return None
        # Retrieve data
        _, simmatLogP, labx = self._get_adjmat(summarize)

        config = {}
        config['scale'] = scale
        config['node_color'] = node_color
        config['dist_between_nodes'] = dist_between_nodes
        config['node_size_limits'] = node_size_limits
        config['layout'] = layout
        config['iterations'] = 50
        config['dpi'] = dpi

        # Set savepath and filename
        config['savepath'] = hnstats._path_correct(savepath, filename='hnet_network', ext='.png')
        # Get adjmat
        # adjmatLog = self.results['simmatLogP'].copy()
        # Filter adjacency matrix on blacklist/whitelist and/or threshold
        adjmatLog, labx = hnstats._filter_adjmat(simmatLogP, labx, threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
        # Check whether anything has remained
        if adjmatLog.values.flatten().sum()==0:
            if verbose>=3: print('[hnet] >Nothing to plot.')
            return None

        # Set weights for edges
        adjmatLogWEIGHT = adjmatLog.copy()
        np.fill_diagonal(adjmatLogWEIGHT.values, 0)
        adjmatLogWEIGHT = pd.DataFrame(index=adjmatLog.index.values, data=MinMaxScaler(feature_range=(0, 20)).fit_transform(adjmatLogWEIGHT), columns=adjmatLog.columns)

        # Set size for node
        IA, IB = ismember(adjmatLog.columns, self.results['counts'][:, 0])
        node_size = np.repeat(node_size_limits[0], len(adjmatLog.columns))
        node_size[IA] = hnstats._scale_weights(self.results['counts'][IB, 1], node_size_limits)

        # Make new graph (G) and add properties to nodes
        if not directed:
            adjmatLog = to_undirected(adjmatLog, method='logp')
            G = nx.Graph()
        else:
            G = nx.DiGraph(directed=True)

        # Color node using network-clustering
        # if node_color=='cluster':
            # _, labx = network.cluster(G.to_undirected())
        # else:
            # labx = self.results['labx']
        # Make sure the labx are correct if previously black/white list filtering was performed
        # IA,_ = ismember(self.results['simmatLogP'].columns, adjmatLog.columns)
        # labx = label_encoder.fit_transform(labx[IA])
        labx = label_encoder.fit_transform(labx)

        for i in range(0, adjmatLog.shape[0]):
            G.add_node(adjmatLog.index.values[i], node_size=node_size[i], node_label=labx[i])
        # Add properties to edges
        np.fill_diagonal(adjmatLog.values, 0)
        for i in range(0, adjmatLog.shape[0]):
            idx=np.where(adjmatLog.iloc[i, :]>0)[0]
            labels=adjmatLog.iloc[i, idx]
            labelsWEIGHT=adjmatLogWEIGHT.iloc[i, idx]

            for k in range(0, len(labels)):
                G.add_edge(labels.index[k], adjmatLog.index[i], weight=labels.values[k].astype(int), capacity=labelsWEIGHT.values[k])

        edges = G.edges()
        edge_weights = np.array([G[u][v]['weight'] for u, v in edges])
        edge_weights = MinMaxScaler(feature_range=(0.5, 8)).fit_transform(edge_weights.reshape(-1, 1)).flatten()

        # # Cluster
        if node_color=='cluster':
            _, labx = network.cluster(G.to_undirected())
        else:
            labx = label_encoder.fit_transform(labx)

        # G = nx.DiGraph() # Directed graph
        # Layout
        if isinstance(pos, type(None)):
            pos = nx.fruchterman_reingold_layout(G, weight='edge_weights', k=config['dist_between_nodes'], scale=config['scale'], iterations=config['iterations'])
        else:
            pos = network.graphlayout(G, pos=pos, scale=config['scale'], layout=config['layout'])

        # pos = nx.spring_layout(G, weight='edge_weights', k=config['dist_between_nodes'], scale=config['scale'], iterations=config['iterations'])
        # Boot-up figure
        if showfig or (not isinstance(config['savepath'], type(None))):
            [fig, ax] = plt.subplots(figsize=figsize)
            options = {
                # 'node_color': 'grey',
                'arrowsize': 12,
                'font_size': 18,
                'font_color': 'black'}
            # Draw plot
            nx.draw(G, pos, with_labels=True, **options, node_size=node_size * 5, width=edge_weights, node_color=labx, cmap='Paired')
            # Plot weights
            nx.draw_networkx_edge_labels(G, pos, edge_labels=nx.get_edge_attributes(G, 'weight'))
            # Figure cleaning
            ax = plt.gca()
            ax.set_axis_off()
            # Show figure
            if showfig: plt.show()
            # Save figure to disk
            if not isinstance(config['savepath'], type(None)):
                savefig(fig, config['savepath'], dpi=config['dpi'], transp=True)

        # Return
        Gout = {}
        Gout['G'] = G
        Gout['labx'] = labx
        Gout['pos'] = pos
        return(Gout)

    # Make plot of the association_learning
    def heatmap(self, summarize=False, cluster=False, figsize=[15, 15], savepath=None, threshold=None, white_list=None, black_list=None, min_edges=None, verbose=3):
        """Plot static heatmap.

        Description
        -----------
        A heatmap can be of use when the results becomes too large to plot in a network.

        Parameters
        ----------
        self : Object
            The output of .association_learning()
        summarize : bool, (default: False)
            Show the results based on categoric or label-specific associations.
            True: Summrize based on the categories
            False: All associations across labels
        cluster : Bool, optional
            Cluster before making heatmap. The default is False.
        figsize : typle, optional
            Figure size. The default is [15, 10].
        savepath : Bool, optional
            saveingpath. The default is None.
        threshold : int (default : None)
            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
        black_list : List or None (default : None)
            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
        white_list : List or None (default : None)
            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
        min_edges : int (default : None)
            Edges are only shown if a node has at least min_edges.
        verbose : int, optional
            Verbosity. The default is 3.

        Returns
        -------
        None.

        """
        if verbose>=3: print('[hnet] >Building static heatmap.')
        status = self._check_results()
        if not status: return None
        # Retrieve data
        _, simmatLogP, labx = self._get_adjmat(summarize)

        # adjmatLog = self.results['simmatLogP'].copy()
        # Filter adjacency matrix on blacklist/whitelist and/or threshold
        adjmatLog, labx = hnstats._filter_adjmat(simmatLogP, labx, threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
        if adjmatLog.values.flatten().sum()==0:
            if verbose>=3: print('[hnet] >Nothing to plot.')
            return None

        # Set savepath and filename
        savepath = hnstats._path_correct(savepath, filename='hnet_heatmap', ext='.png')

        try:
            savepath1=''
            if savepath is not None:
                [getdir, getname, getext]=hnstats._path_split(savepath)
                if getname=='': getname='heatmap'
                savepath1 = os.path.join(getdir, getname + '_logP' + getext)

            if cluster:
                fig1=imagesc.cluster(adjmatLog.fillna(value=0).values, row_labels=adjmatLog.index.values, col_labels=adjmatLog.columns.values, cmap='Reds', figsize=figsize)
            else:
                fig1=imagesc.plot(adjmatLog.fillna(value=0).values, row_labels=adjmatLog.index.values, col_labels=adjmatLog.columns.values, cmap='Reds', figsize=figsize)

            if savepath is not None:
                if verbose>=3: print('[hnet] >Saving heatmap..')
                _ = savefig(fig1, savepath1, transp=True)
        except:
            print('[hnet] >Error: Heatmap failed. Try cluster=False')

    # Extract combined rules from association_learning
    def combined_rules(self, simmatP=None, labx=None, verbose=3):
        """Association testing and combining Pvalues using fishers-method.

        Description
        -----------
        Multiple variables (antecedents) can be associated to a single variable (consequent).
        To test the significance of combined associations we used fishers-method. The strongest connection will be sorted on top.

        Parameters
        ----------
        simmatP : matrix
            simmilarity matrix
        verbose : int, optional
            Print message to screen. The higher the number, the more details. The default is 3.

        Returns
        -------
        pd.DataFrame()
            Dataset containing antecedents and consequents. The strongest connection will be sorted on top.
            The columns are as following:
        antecedents_labx
            Generic label name.
        antecedents
            Specific label names in the 'from' category.
        consequents
            Specific label names that are the result of the antecedents.
        Pfisher
            Combined P-value

        Examples
        --------
        >>> from hnet import hnet
        >>> hn = hnet()
        >>> df = hn.import_example('sprinkler')
        >>> hn.association_learning(df)
        >>> hn.combined_rules()
        >>> print(hn.rules)

        """
        if simmatP is None:
            if self.results.get('simmatP', None) is None: raise ValueError('[hnet] >Error: Input requires the result from the association_learning() function.')
            simmatP = self.results['simmatP']
        if labx is None:
            labx = self.results['labx']

        df_rules = pd.DataFrame(index=np.arange(0, simmatP.shape[0]), columns=['antecedents_labx', 'antecedents', 'consequents', 'Pfisher'])
        df_rules['consequents'] = simmatP.index.values

        for i in tqdm(range(0, simmatP.shape[0]), disable=(True if verbose==0 else False)):
            idx = np.where(simmatP.iloc[i, :]<1)[0]
            # Remove self
            idx = np.setdiff1d(idx, i)
            # Store rules
            df_rules['antecedents'].iloc[i] = list(simmatP.iloc[i, idx].index)
            df_rules['antecedents_labx'].iloc[i] = labx[idx]
            # Combine pvalues
            df_rules['Pfisher'].iloc[i] = combine_pvalues(simmatP.iloc[i, idx].values, method='fisher')[1]

        # Keep only lines with pvalues
        df_rules.dropna(how='any', subset=['Pfisher'], inplace=True)
        # Sort
        df_rules.sort_values(by=['Pfisher'], ascending=True, inplace=True)
        df_rules.reset_index(inplace=True, drop=True)
        # Return
        return(df_rules)

    def import_example(self, data='titanic', url=None, sep=',', verbose=3):
        """Import example dataset from github source.

        Description
        -----------
        Import one of the few datasets from github source or specify your own download url link.

        Parameters
        ----------
        data : str
            Name of datasets: 'sprinkler', 'titanic', 'student', 'fifa', 'cancer', 'waterpump', 'retail'
        url : str
            url link to to dataset.
        verbose : int, (default: 3)
            Print message to screen.

        Returns
        -------
        pd.DataFrame()
            Dataset containing mixed features.

        """
        return import_example(data=data, url=url, sep=sep, verbose=verbose)

    # Save model
    def save(self, filepath='hnet_model.pkl', overwrite=False, verbose=3):
        """Save learned model in pickle file.

        Parameters
        ----------
        filepath : str, (default: 'hnet_model.pkl')
            Pathname to store pickle files.
        overwrite : bool, (default=False)
            Overwite file if exists.
        verbose : int, optional
            Show message. A higher number gives more informatie. The default is 3.

        Returns
        -------
        bool : [True, False]
            Status whether the file is saved.

        """
        if (filepath is None) or (filepath==''):
            filepath = 'hnet_model.pkl'
        if filepath[-4:] != '.pkl':
            filepath = filepath + '.pkl'
        # Store data
        storedata = {}
        storedata['results'] = self.results
        storedata['alpha'] = self.alpha
        storedata['k'] = self.k
        storedata['y_min'] = self.y_min
        storedata['multtest'] = self.multtest
        storedata['dtypes'] = self.dtypes
        storedata['specificity'] = self.specificity
        storedata['perc_min_num'] = self.perc_min_num
        storedata['dropna'] = self.dropna
        storedata['fillna'] = self.fillna
        storedata['excl_background'] = self.excl_background
        # Save
        status = pypickle.save(filepath, storedata, overwrite=overwrite, verbose=verbose)
        if verbose>=3: print('[hnet] >Saving.. %s' %(status))
        # return
        return status

    # Load model.
    def load(self, filepath='hnet_model.pkl', verbose=3):
        """Load learned model.

        Parameters
        ----------
        filepath : str
            Pathname to stored pickle files.
        verbose : int, optional
            Show message. A higher number gives more information. The default is 3.

        Returns
        -------
        Object.

        """
        if (filepath is None) or (filepath==''):
            filepath = 'hnet_model.pkl'
        if filepath[-4:]!='.pkl':
            filepath = filepath + '.pkl'
        # Load
        storedata = pypickle.load(filepath, verbose=verbose)
        # Store in self.
        if storedata is not None:
            self.results = storedata['results']
            self.alpha = storedata['alpha']
            self.k = storedata['k']
            self.y_min = storedata['y_min']
            self.multtest = storedata['multtest']
            self.dtypes = storedata['dtypes']
            self.specificity = storedata['specificity']
            self.perc_min_num = storedata['perc_min_num']
            self.dropna = storedata['dropna']
            self.fillna = storedata['fillna']
            self.excl_background = storedata['excl_background']
            if verbose>=3: print('[hnet] >Loading succesfull!')
            # Return results
            return self.results
        else:
            if verbose>=2: print('[hnet] >WARNING: Could not load data.')

    # Check results
    def _check_results(self, verbose=3):
        status = True
        if not hasattr(self, 'results'):
            if verbose>=3: print('[hnet] >Nothing to plot. Try to run hnet with first with: hn.association_learning()')
            status = False
        elif self.results['simmatLogP'].empty:
            if verbose>=3: print('[hnet] >Nothing to plot. No associations were detected.')
            status = False

        return status


# %% Store results
def _store(simmatP, adjmatLog, GsimmatP, GsimmatLogP, labx, df, nr_succes_pop_n, dtypes, rules, feat_importance):
    out = {}
    out['simmatP'] = simmatP
    out['simmatLogP'] = adjmatLog
    out['simmatP_cat'] = GsimmatP
    out['simmatLogP_cat'] = GsimmatLogP
    out['labx'] = labx.astype(str)
    out['dtypes'] = np.array(list(zip(df.columns.values.astype(str), dtypes)))
    out['counts'] = nr_succes_pop_n
    out['rules'] = rules
    out['feat_importance'] = feat_importance
    return out


# %% Import example dataset from github.
def import_example(data='titanic', url=None, sep=',', verbose=3):
    """Import example dataset from github source.

    Description
    -----------
    Import one of the few datasets from github source or specify your own download url link.

    Parameters
    ----------
    data : str
        Name of datasets: 'sprinkler', 'titanic', 'student', 'fifa', 'cancer', 'waterpump', 'retail'
    url : str
        url link to to dataset.
    verbose : int, (default: 3)
        Print message to screen.

    Returns
    -------
    pd.DataFrame()
        Dataset containing mixed features.

    """
    if url is None:
        if data=='sprinkler':
            url='https://erdogant.github.io/datasets/sprinkler.zip'
        elif data=='titanic':
            url='https://erdogant.github.io/datasets/titanic_train.zip'
        elif data=='student':
            url='https://erdogant.github.io/datasets/student_train.zip'
        elif data=='cancer':
            url='https://erdogant.github.io/datasets/cancer_dataset.zip'
        elif data=='fifa':
            url='https://erdogant.github.io/datasets/FIFA_2018.zip'
        elif data=='waterpump':
            url='https://erdogant.github.io/datasets/waterpump/waterpump_test.zip'
        elif data=='retail':
            url='https://erdogant.github.io/datasets/marketing_data_online_retail_small.zip'
    else:
        data = wget.filename_from_url(url)

    if url is None:
        if verbose>=3: print('[hnet] >Nothing to download.')
        return None

    curpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
    PATH_TO_DATA = os.path.join(curpath, wget.filename_from_url(url))
    if not os.path.isdir(curpath):
        os.makedirs(curpath, exist_ok=True)

    # Check file exists.
    if not os.path.isfile(PATH_TO_DATA):
        if verbose>=3: print('[hnet] >Downloading [%s] dataset from github source..' %(data))
        wget.download(url, curpath)

    # Import local dataset
    if verbose>=3: print('[hnet] >Import dataset [%s]' %(data))
    df = pd.read_csv(PATH_TO_DATA, sep=sep)
    # Return
    return df


# %% Compute fit
def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas', specificity='medium', excl_background=None, verbose=3):
    """Enrichment analysis.

    Description
    -----------
    Compute enrichment between input dataset and response variable y. Length of dataframe and y must be equal.
    The input dataset is converted into a one-hot dense array based on automatic typing ``dtypes='pandas'`` or user defined dtypes.

    Parameters
    ----------
    df : DataFrame
        Input Dataframe.
    y : list of length df.index
        Response variable.
    y_min : int, optional
        Minimal number of samples in a group.. The default is None.
    alpha : float, optional
        Significance. The default is 0.05.
    multtest : String, optional
        Multiple test correcton. The default is 'holm'.
    dtypes : list of length df.columns, optional
        By default the dtype is determined based on the pandas dataframe. Empty ones [''] are skipped. The default is 'pandas'.
    specificity : String, optional
        Configure how numerical data labels are stored.. The default is 'medium'.
    excl_background : String (default : None)
        Name to exclude from the background.
        Example: ['0.0']: To remove categorical values with label 0
    verbose : int, optional
        Print message to screen. The higher the number, the more details. The default is 3.

    Returns
    -------
    pd.DataFrame() with the following columns:
    category_label : str
        Label of the category.
    P :  float
        Pvalue of the hypergeometric test or Wilcoxon Ranksum.
    logP :  float
        -log10(Pvalue) of the hypergeometric test or Wilcoxon Ranksum.
    Padj :  float
        Adjusted P-value.
    dtype : list of str
        Categoric or numeric.
    y : str
        Response variable name.
    category_name : str
        Subname of the category_label.
    popsize_M : int
        Population size: Total number of samples.
    nr_succes_pop_n : int
        Number of successes in population.
    overlap_X : int
        Overlap between response variable y and input feature.
    samplesize_N : int
        Sample size: Random variate, eg clustersize or groupsize, those of interest.
    zscore : float
        Z-score of the Wilcoxon Ranksum test.
    nr_not_succes_pop_n : int
        Number of successes in population.

    Examples
    --------
    >>> import hnet
    >>> df = hn.import_example('titanic')
    >>> y = df['Survived'].values
    >>> out = hnet.enrichment(df, y)

    """
    assert isinstance(df, pd.DataFrame), 'Data must be of type pd.DataFrame()'
    assert len(y)==df.shape[0], 'Length of [df] and [y] must be equal'
    assert 'numpy' in str(type(y)), 'y must be of type numpy array'

    # DECLARATIONS
    config = {}
    config['verbose'] = verbose
    config['alpha'] = alpha
    config['multtest'] = multtest
    config['specificity'] = specificity

    if config['verbose']>=3: print('[hnet] >Start making fit..')
    df.columns = df.columns.astype(str)

    # [df, df_onehot, dtypes] = hnstats._preprocessing(df, dtypes=dtypes, y_min=y_min, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
    df, dtypes, excl_background = hnstats._bool_processesing(df, dtypes, excl_background=excl_background, verbose=verbose)

    # Set y as string
    y = df2onehot.set_y(y, y_min=y_min, verbose=config['verbose'])
    # Determine dtypes for columns
    df, dtypes = df2onehot.set_dtypes(df, dtypes, verbose=config['verbose'])
    # Compute fit
    out = hnstats._compute_significance(df, y, dtypes, specificity=config['specificity'], verbose=config['verbose'])
    # Multiple test correction
    out = hnstats._multipletestcorrection(out, config['multtest'], verbose=config['verbose'])
    # Keep only significant ones
    out = hnstats._filter_significance(out, config['alpha'], multtest)
    # Make dataframe
    out = pd.DataFrame(out)
    # Return
    if config['verbose']>=3: print('[hnet] >Fin')
    return(out)


# %% Make adjacency matrix symmetric with repect to the diagonal
def to_undirected(adjmat, method='logp', verbose=3):
    """Make adjacency matrix symmetric.

    Description
    -----------
    The adjacency matrix resulting from hnet is not neccesarily symmetric due to the statistics being used.
    In some cases, a symmetric matrix can be usefull. This function makes sure that values above the diagonal are the same as below the diagonal.
    Values above and below the diagnal are combined using the max or min value.

    Parameters
    ----------
    adjmat : array
        Square form adjacency matrix.
    method : str
        Make matrix symmetric using the 'max' or 'min' function.
    verbose : int
        Verbosity. The default is 3.

    Returns
    -------
    pd.DataFrame().
        Symmetric adjacency matrix.

    """
    if np.sum(np.isin(adjmat.index.values, adjmat.columns.values))!=np.max(adjmat.shape):
        raise ValueError('Adjacency matrix must have similar number of rows and columns! Re-run HNet with dropna=False!')

    if verbose>=3: print('[hnet] >Make adjacency matrix undirected..')
    progressbar=(True if verbose==0 else False)
    columns=adjmat.columns.values
    index=adjmat.index.values

    # Make selected matrix symmetric
    if isinstance(adjmat, pd.DataFrame):
        adjmat=adjmat.values

    # Make output matrix
    adjmatS=np.zeros(adjmat.shape, dtype=float)

    # Make symmetric using maximum as combining function
    for i in tqdm(range(adjmat.shape[0]), disable=progressbar):
        for j in range(adjmat.shape[1]):
            if method=='logp':
                # Maximum -log10(P) values
                score = np.maximum(adjmat[i, j], adjmat[j, i])
            else:
                # Minimum P-values
                score = np.minimum(adjmat[i, j], adjmat[j, i])

            adjmatS[i, j] = score
            adjmatS[j, i] = score

    # Make dataframe and return
    adjmatS=pd.DataFrame(index=index, data=adjmatS, columns=columns, dtype=float)
    # Return
    return(adjmatS)


# %% Comparison of two networks
def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15, height=8, verbose=3):
    """Compare two adjacency matrices and plot the differences.

    Description
    -----------
    Comparison of two networks based on two adjacency matrices. Both matrices should be of equal size and of type pandas DataFrame.
    The columns and rows between both matrices are matched if not ordered similarly.

    Parameters
    ----------
    adjmat_true : pd.DataFrame()
        First array.
    adjmat_pred : pd.DataFrame()
        Second array.
    pos : dict, optional
        Position of the nodes. The default is None.
    showfig : Bool, optional
        Plot figure. The default is True.
    width : int, optional
        Width of the figure. The default is 15.
    height : int, optional
        Height of the figure. The default is 8.
    verbose : int, optional
        Verbosity. The default is 3.

    Returns
    -------
    tuple
        Output contains a tuple of two elements, the score of matching adjacency matrix and adjacency matrix differences.
    scores : dict
        Contains extensive number of keys with various scoring values.
    adjmat_diff : pd.DataFrame()
        Difference of the input network compared to the second.
        0 = No edge
        1 = No difference between networks
        2 = Addition of edge in the first input network compared to the second
       -1 = Depliction of edge in the first netwwork compared to the second

    """
    scores, adjmat_diff = network.compare_networks(adjmat_true,
                                                   adjmat_pred,
                                                   pos=pos,
                                                   showfig=showfig,
                                                   width=width,
                                                   height=height,
                                                   verbose=verbose)
    return(scores, adjmat_diff)


# %% Do the math
def _do_the_math(df, X_comb, dtypes, X_labx, simmatP, simmat_labx, i, specificity, y_min, verbose=3):
    count = 0
    # Get response variable to test association
    y = X_comb.iloc[:, i].values.astype(str)
    # Get column name
    colname = X_comb.columns[i]
    # Default output is nan
    out = [colname, np.nan]
    # Do math if response variable has more then 1 option
    if len(np.unique(y))>1:
        if verbose>=4: print('[hnet] >Working on [%s]' %(X_comb.columns[i]), end='')
        # Remove columns if it belongs to the same categorical subgroup; these can never overlap!
        Iloc = ~np.isin(df.columns, X_labx[i])
        # Compute fit
        dfout = enrichment(df.loc[:, Iloc], y, y_min=y_min, alpha=1, multtest=None, dtypes=dtypes[Iloc], specificity=specificity, verbose=0)
        # Count
        count=count + dfout.shape[0]
        # Match with dataframe and store
        if not dfout.empty:
            # Column names
            idx = np.where(dfout['category_label'].isna())[0]
            catnames = dfout['category_name']
            colnames = catnames + '_' + dfout['category_label'].astype(str)
            colnames[idx] = catnames[idx].values
            # Add new column and index
            simmatP, simmat_labx = hnstats._addcolumns(simmatP, colnames, simmat_labx, catnames)
            # Store values
            IA, IB = ismember(simmatP.index.values.astype(str), colnames.values.astype(str))
            simmatP.loc[colname, IA] = dfout['Padj'].iloc[IB].values
            # Count nr. successes
            out = [colname, X_comb.iloc[:, i].sum() / X_comb.shape[0]]
            # showprogress
            if verbose>=4: print('[%g]' %(len(IB)), end='')
    else:
        if verbose>=4: print('[hnet] >Skipping [%s] because length of unique values=1' %(X_comb.columns[i]), end='')

    if verbose>=4: print('')
    # Return
    return(out, simmatP, simmat_labx)
